class CategorySimilarityMatrix(object):
    """ Similarity measure for categorical attributes defined manually
    """
    def compute_similarity(self, col, val1, val2, aggr_col):
        # if col == 'pubkey':
            # print('In compute_similarity:', col, val1, val2, self.sim_matrix[col][val1])
        if col not in self.sim_matrix:
            return 0
        else:
            if val1 not in self.sim_matrix[col] and val2 not in self.sim_matrix[col]:
                if val1 == val2:
                    return 1
                else:
                    return 0
            return self.sim_matrix[col][val1][val2]

    def __init__(self, inf='', schema=None):
        self.sim_matrix = {}
        if schema is not None:
            for k in schema:
                if schema[k] != 'integer' and not schema[k].startswith('double') and not schema[k].startswith('float'):
                    self.sim_matrix[k] = {}
        if inf == '':
            return
        infile = open(inf, 'r')

        col = ''
        while True:
            line = infile.readline()
            if not line:
                break
            temp_arr = line.split(',')
            if len(temp_arr) == 1:
                col = line.strip()
                self.sim_matrix[col] = {}
            else:
                sim_tuple = temp_arr
                if sim_tuple[0] not in self.sim_matrix[col]:
                    self.sim_matrix[col][sim_tuple[0]] = {sim_tuple[0]:1.0}
                if sim_tuple[1] not in self.sim_matrix[col]:
                    self.sim_matrix[col][sim_tuple[1]] = {sim_tuple[1]:1.0}
                self.sim_matrix[col][sim_tuple[0]][sim_tuple[1]] = float(sim_tuple[2])
                self.sim_matrix[col][sim_tuple[1]][sim_tuple[0]] = float(sim_tuple[2])
        print('Similarity Matrix')
        print(self.sim_matrix)


    def is_categorical(self, col):
        return col in self.sim_matrix


#!/usr/bin/python
# -*- coding:utf-8 -*- 

import psycopg2
import sys, getopt
import pandas
import csv
#import statsmodels.formula.api as smf
from sklearn import preprocessing
import math
import time
from heapq import *
import re
import itertools

from capexplain.similarity.category_similarity_matrix import *
from capexplain.similarity.category_network_embedding import *
from capexplain.similarity.category_similarity_naive import *
from capexplain.similarity.category_similarity_matrix import *
from capexplain.similarity.category_network_embedding import *
from capexplain.utils import *
from capexplain.pattern_model.LocalRegressionPattern import *


DEFAULT_QUERY_RESULT_TABLE = 'crime_2017_2'
# DEFAULT_PATTERN_TABLE = 'pub_select'
DEFAULT_PATTERN_TABLE = 'crime_2017_2'
DEFAULT_USER_QUESTION_NUMBER = 5
EXAMPLE_NETWORK_EMBEDDING_PATH = './input/NETWORK_EMBEDDING'
EXAMPLE_SIMILARITY_MATRIX_PATH = './input/SIMILARITY_DEFINITION'
DEFAULT_AGGREGATE_COLUMN = '*'
DEFAULT_EPSILON = 0.25
TOP_K = 5

def build_local_regression_constraint(data, column_index, t, con, epsilon, agg_col, regression_package):
    """Build local regression constraint from Q(R), t, and global regression constraint

    Args:
        data: result of Q(R)
        column_index: index for values in each column
        t: target tuple in Q(R)
        con: con[0] is the list of fixed attributes in Q(R), con[1] is the list of variable attributes in Q(R)
        epsilon: threshold for local regression constraint
        regression_package: which package is used to compute regression 
    Returns:
        A LocalRegressionConstraint object whose model is trained on \pi_{con[1]}(Q_{t[con[0]]}(R))
    """
    tF = get_F_value(con[0], t)
    local_con = LocalRegressionConstraint(con[0], tF, con[1], agg_col, epsilon)
    train_data = {agg_col: []}
    for v in con[1]:
        train_data[v] = []
    # for index, row in data['df'].iterrows():
    #     if get_F_value(con[0], row) == tF:
    #         for v in con[1]:
    #             train_data[v].append(row[v])
    #         train_data[agg_col].append(row[agg_col])

    
    for idx in column_index[con[0][0]][tF[0]]:
        row = data['df'].loc[data['df']['index'] == idx]
        row = row.to_dict('records')[0]
        #print row
        if get_F_value(con[0], row) == tF:
            for v in con[1]:
                train_data[v].append(row[v])
            train_data[agg_col].append(row[agg_col])
    if regression_package == 'scikit-learn':
        train_x = {}
        for v in con[1]:
            if v in data['le']:
                train_data[v] = data['le'][v].transform(train_data[v])
                train_data[v] = data['ohe'][v].transform(train_data[v].reshape(-1, 1))
                #print data['ohe'][v].transform(train_data[v].reshape(-1, 1))
                train_x[v] = train_data[v]
            else:
                if v != agg_col:
                    train_x[v] = np.array(train_data[v]).reshape(-1, 1)
        train_y = np.array(train_data[agg_col]).reshape(-1, 1)
        train_x = np.concatenate(list(train_x.values()), axis=-1)
        local_con.train_sklearn(train_x, train_y)
    else:
        #train_data = pandas.DataFrame(train_data)
        formula = agg_col + ' ~ ' + ' + '.join(con[1])
        local_con.train(train_data, formula)
    return local_con

def predict(local_pattern, t):
    # print('In predict ', local_pattern)
    if local_pattern[5] == 'const':
        # predictY = float(local_pattern[-1][1:-1])
        predictY = float(local_pattern[-2][1:-1].split(',')[0])
    elif local_pattern[5] == 'linear':
        # print(local_pattern, t)
        v = get_V_value(local_pattern[2], t)
        # params = list(map(float, local_pattern[-1][1:-1].split(',')))
        params_str = local_pattern[-1].split('\n')
        # params = list(map(float, ))
        # print(params_str, v)
        params_dict = {}
        for i in range(0, len(params_str)-1):
            
            # p_cate = re.compile(r'(.*)\[T\.\s+(.*)\]\s+(-?\d+\.\d+)')
            p_cate = re.compile(r'(.*)\[T\.\s*(.*)\]\s+(-?\d+\.\d+)')
            cate_res = p_cate.findall(params_str[i])
            #print(params_str[i], cate_res)
            if len(cate_res) != 0:
                cate_res = cate_res[0]
                v_attr = cate_res[0]
                v_val = cate_res[1]
                param = float(cate_res[2])
                if v_attr not in params_dict:
                    params_dict[v_attr] = {}
                params_dict[v_attr][v_val] = param
            else:
                p_nume = re.compile(r'([^\s]+)\s+(-?\d+\.\d+)')
                nume_res = p_nume.findall(params_str[i])
                if len(nume_res) == 0:
                    continue
                # print(nume_res)
                v_attr = nume_res[0][0]
                param = float(nume_res[0][1])
                params_dict[v_attr] = param

        predictY = 0.0
        # print(params_dict)
        for v_attr, v_dict in params_dict.items():
            # print(v_attr, v_dict, t)
            if v_attr == 'Intercept':
                predictY += v_dict
            else:
                if isinstance(v_dict, dict):
                    # print(t, v_attr, v_dict)
                    v_key = t[v_attr].replace('\'', '').replace(' ', '')
                    # print(v_attr, v_key)
                    # print(v_dict.keys())
                    if v_key in v_dict:
                        predictY += v_dict[v_key]
                else:
                    if v_attr in t:
                        predictY += v_dict * t[v_attr]

        # predictY = sum(map(lambda x: x[0]*x[1], zip(params[:-1], v))) + params[-1]

    return predictY

def validate_local_regression_pattern(local_pattern, epsilon, t, dir, agg_col, cur, table_name):
    """Check the validicity of the user question under a local regression constraint

    Args:
        local_pattern:
        t: target tuple in Q(R)
        dir: whether user thinks t[agg(B)] is high or low
        agg_col: the column of aggregated value
    Returns:
        the actual direction that t[agg(B)] compares to its expected value, and the expected value from local_con
    """
    print(local_pattern, t)
    test_tuple = {}
    # print('PAT', local_pattern)
    if isinstance(local_pattern, dict):
        return -2, 0
    for v in local_pattern[2]:
        test_tuple[v] = t[v.replace(' ', '')]
    # if regression_package == 'scikit-learn':
    #     for v in local_con.var_attr:
    #         if v in data['le']:
    #             test_tuple[v] = data['le'][v].transform(test_tuple[v])
    #             test_tuple[v] = data['ohe'][v].transform(test_tuple[v].reshape(-1, 1))
    #         else:
    #             test_tuple[v] = np.array(test_tuple[v]).reshape(-1, 1)
        
    #     test_tuple = np.concatenate(list(test_tuple.values()), axis=-1)
    #     predictY = local_con.predict_sklearn(test_tuple)
    # else:
    #     predictY = local_con.predict(pandas.DataFrame(test_tuple))

    predictY = predict(local_pattern, test_tuple)
    if t[agg_col] < (1-epsilon) * predictY:
        # print(test_tuple, predictY)
        return -dir, predictY
    elif t[agg_col] > (1+epsilon) * predictY:
        # print(test_tuple, predictY)
        return dir, predictY
    else:
        return 0, predictY
        
def tuple_similarity(t1, t2, var_attr, cat_sim, num_dis_norm, agg_col):
    """Compute the similarity between two tuples t1 and t2 on their attributes var_attr

    Args:
        t1, t2: two tuples
        var_attr: variable attributes
        cat_sim: the similarity measure for categorical attributes
        num_dis_norm: normalization terms for numerical attributes
        agg_col: the column of aggregated value
    Returns:
        the Gower similarity between t1 and t2
    """
    sim = 0.0
    cnt = 0
    for v_col in var_attr:
        col = v_col.replace(' ', '')
        
        if t1[col] is None or t2[col] is None:
            continue
        if cat_sim.is_categorical(col):
            t1_key = t1[col].replace("'", '').replace(' ', '')
            t2_key = t2[col].replace("'", '').replace(' ', '')
            s = cat_sim.compute_similarity(col, t1_key, t2_key, agg_col)
            # print(s)
            sim += s
        else:
            print(col, t1, t2, num_dis_norm[col])
            # print( num_dis_norm[col]['range'])
            if num_dis_norm[col]['range'] is None:
                if t1[col] == t2[col]:
                    sim += 1
            else:
                if col != agg_col and col != 'index':
                    temp = abs(t1[col] - t2[col]) / num_dis_norm[col]['range']
                    sim += 1-temp
        cnt += 1
    # print(t1, t2, sim)
    return sim / cnt

def get_local_pattern(g_pat, t, cur, table_name):
    tF = get_F_value(g_pat[0], t)
    # print(tF)
    local_pattern_query = '''SELECT * FROM {} WHERE REPLACE(fixed_value, ' ', '')=REPLACE('{}', ' ', '') AND variable='{}' AND in_a='{}' AND agg='{}' AND model='{}';'''.format(
                table_name + '_local', str(tF).replace("\'", ''), 
                str(g_pat[1]).replace("\'", ''), g_pat[2], g_pat[3], g_pat[4]
        )
    # print(local_pattern_query)
    cur.execute(local_pattern_query)
    res = cur.fetchall()
    l_pat = []
    # print('In get local patters: ', res)
    if len(res) != 0:
        for i in range(3):
            l_pat.append(res[0][i][1:-1].split(','))
        l_pat[0] = list(map(lambda x: x.replace(' ', ''), l_pat[0]))
        l_pat[2] = list(map(lambda x: x.replace(' ', ''), l_pat[2]))
        for i in range(3, len(res[0])):
            l_pat.append(res[0][i])
        # print(l_pat)
        return l_pat
    else:
        l_pat.append(g_pat[0])
        l_pat.append(tF)
        for i in range(1, len(g_pat)):
            l_pat.append(g_pat[i])
        return {'hold': False, 'pattern': l_pat}

def get_local_patterns(global_patterns, t, cur, table_name):
    local_patterns = []
    local_patterns_dict = {}
    for pat in global_patterns:
        
        F_set = set(pat[0])
        V_set = set(pat[1])
        t_set = set(t.keys())
        FV_set = F_set.union(V_set)
        if not (FV_set.issubset(t_set) and len(FV_set) == len(t_set) - 1 
            and '{}({})'.format(pat[3], pat[2]) in t):
        # if F_set.union(V_set) != t_set:
            continue
        # print(pat[0], t)
        local_patterns.append(get_local_pattern(pat, t, cur, table_name))


    # for num_of_fix in range(1, min(len(t.keys()),4)):
    #     for possible_f_attr in itertools.combinations(list(t.keys())[:-1], num_of_fix):
    #         tF = get_F_value(possible_f_attr, t)
    #         local_pattern_query = '''SELECT * FROM {} WHERE REPLACE(fixed_value, ' ', '')=REPLACE('{}', ' ', '')'''.format(
    #             table_name + '_local', str(tF).replace("\'", '"'))
    #         # print(tF)
    #         print(local_pattern_query)
    #         cur.execute(local_pattern_query)
    #         res = cur.fetchall()
    #         # print('In get local patterns: ', res)
    #         for k in range(len(res)):
    #             l_pat = []
    #             if str(res[k]) in local_patterns_dict:
    #                 continue
    #             else:
    #                 local_patterns_dict[str(res[k])] = True
    #             for i in range(3):
    #                 l_pat.append(res[k][i][1:-1].split(','))
    #             l_pat[0] = list(map(lambda x: x.replace(' ', ''), l_pat[0]))
    #             l_pat[2] = list(map(lambda x: x.replace(' ', ''), l_pat[2]))
    #             for i in range(3, len(res[0])):
    #                 l_pat.append(res[0][i])
    #             # print(l_pat)
    #             if l_pat not in local_patterns:
    #                 local_patterns.append(l_pat)
    # print('GET ', local_patterns)
    return local_patterns


def get_tuples_by_F(local_pattern, f_value, cur, table_name, cat_sim):
    def tuple_column_to_str_in_where_clause_2(col_value):
        # print(col_value, cat_sim.is_categorical(col_value[0]))
        if cat_sim.is_categorical(col_value[0]):
            return "like '%" + str(col_value[1]) + "%'"
        else:
            if is_float(col_value[1]):
                return '=' + str(col_value[1])
            else:
                return "like '%" + str(col_value[1]) + "%'"


    F = str(local_pattern[0]).replace("\'", '')[1:-1]
    V = str(local_pattern[2]).replace("\'", '')[1:-1]
    F_list = F.split(', ')
    V_list = V.split(', ')
    where_clause = ' AND '.join(list(map(lambda x: "{} {}".format(x[0], x[1]), zip(local_pattern[0], map(tuple_column_to_str_in_where_clause_2, zip(F_list, f_value))))))
    tuples_query = '''SELECT {},{},{}({}) FROM {} WHERE {} GROUP BY {}, {};'''.format(
            F, V, local_pattern[4], local_pattern[3], table_name, where_clause, F, V
        )
    # tuples_query = "SELECT * FROM {} WHERE {};".format(table_name, where_clause)
    
    # column_name_query = "SELECT column_name FROM information_schema.columns where table_name='{}';".format(table_name)
    # print(column_name_query)
    # cur.execute(column_name_query)
    # column_name = cur.fetchall()
    column_name = F_list + V_list + [local_pattern[4] + '(' + local_pattern[3] + ')']
    cur.execute(tuples_query)
    # print(tuples_query)
    tuples = []
    res = cur.fetchall()
    min_agg = 1e10
    max_agg = -1e10
    for row in res:
        min_agg = min(min_agg, row[-1])
        max_agg = max(max_agg, row[-1])
        tuples.append(dict(zip(map(lambda x: x, column_name), row)))
        # row_data = {}
        # cnt = 0 
        # for f_attr in F_list:    
        #     if is_float(row[cnt]):
        #         row_data[f_attr] = float(row[cnt])
        #     elif is_integer(row[cnt]):
        #         row_data[f_attr] = float(int(row[cnt]))
        #     else:
        #         row_data[f_attr] = t[cnt]
        #     cnt += 1
        # for v_attr in V_list:
        #     if is_float(row[cnt]):
        #         row_data[v_attr] = float(row[cnt])
        #     elif is_integer(row[cnt]):
        #         row_data[v_attr] = float(int(row[cnt]))
        #     else:
        #         row_data[v_attr] = row[cnt]
        #     cnt += 1
        # row_data[pat[4] + '(' + pat[3] + ')'] = row[-1]
        # tuples.append(row_data)
    return tuples, max_agg - min_agg


def find_explanation_regression_based(user_question_list, global_patterns, cat_sim, num_dis_norm, epsilon, cur, pat_table_name, res_table_name):

    """Find explanations for user questions

    Args:
        data: data['df'] is the data frame storing Q(R)
            data['le'] is the label encoder, data['ohe'] is the one-hot encoder
        user_question_list: list of user questions (t, dir), all questions have the same Q(R)
        cons: list of fixed attributes and variable attributes of global constraints
        cat_sim: the similarity measure for categorical attributes
        num_dis_norm: normalization terms for numerical attributes
        cons_epsilon: threshold for local regression constraints
        agg_col: the column of aggregated value
        regression_package: which package is used to compute regression 
    Returns:
        the top-k list of explanations for each user question
    """

    answer = [[] for i in range(len(user_question_list))]
    local_pattern_loading_time = 0
    question_validating_time = 0
    score_computing_time = 0
    result_merging_time = 0

    local_patterns_list = []


    for j, uq in enumerate(user_question_list):
        dir = uq['dir']
        t = uq['target_tuple']
        print(uq)

        start = time.clock()
        local_patterns = get_local_patterns(global_patterns, t, cur, res_table_name)
        end = time.clock()
        local_pattern_loading_time += end - start

        candidate_list = [[] for i in range(len(local_patterns))]
        top_k_lists = [[] for i in range(len(local_patterns))]
        validate_res_list = []
        local_patterns_list.append(local_patterns)
        
        psi = []
        

        explanation_type = 0
        for i in range(0, len(local_patterns)):
            psi.append(0)
            start = time.clock()
            # print('PAT', i, local_patterns[i])
            if isinstance(local_patterns[i], list):
                agg_col = local_patterns[i][4] + '(' + local_patterns[i][3] + ')'
            else:
                agg_col = 'count(*)'
            validate_res, predicted_aggr = validate_local_regression_pattern(local_patterns[i], epsilon, 
                t, dir, agg_col, 
                cur, res_table_name
            )
            validate_res_list.append(validate_res)
            end = time.clock()
            question_validating_time += end - start

            if validate_res < -1:
                print ("The local regression constraint derived from the " + str(i+1) +
                        "th constraint and the target tuple does not hold")
                explanation_type = 1
                top_k_lists[i] = [1, local_patterns[i]]
            elif validate_res == -1:
                print ("The user question is invalid according to the local regression constraint derived from the "
                        + str(i+1) + "th constraint and the target tuple -- in the opposite direction")
                explanation_type = 2
                top_k_lists[i] = [-2, local_patterns[i], t]
            elif validate_res == 0:
                print ("The user question is invalid according to the local regression constraint derived from the "
                        + str(i+1) + "th constraint and the target tuple -- "
                        "the target tuple can be predicted by the constraint")
                explanation_type = 2
                top_k_lists[i] = [2, local_patterns[i], t]
            else:
                print ("The user question is valid according to the local regression constraint derived from the "
                        + str(i+1) + "th constraint and the target tuple")
                explanation_type = 3
                #deviation = dir * (t[agg_col] - predicted_aggr) / predicted_aggr
                cnt = 0
                #for index, row in data['df'].iterrows():
                start = time.clock()
                f_value = get_F_value(local_patterns[i][0], t)

            ### replace with db query
                tuples_same_F_value, agg_range = get_tuples_by_F(local_patterns[i], f_value, cur, res_table_name, cat_sim)
                max_pred_aggr = -1e10
                min_pred_aggr = 1e10
                for row in tuples_same_F_value:
                    # print(row)
                    psi[-1] += row[agg_col]
                    t_sim = tuple_similarity(t, row, local_patterns[i][2], cat_sim, num_dis_norm, agg_col)
                    # print local_cons[i].var_attr, row
                    test_tuple = dict(zip(local_patterns[i][2],
                                          map(lambda x:x, get_V_value(local_patterns[i][2], row))))
                    predicted_aggr = []
                    
                
                    predicted_aggr = predict(local_patterns[i], test_tuple)
                    if abs(predicted_aggr) < 1e-8:
                        # print(local_patterns[i], test_tuple)
                        deviation = 0
                    else:
                        deviation = (row[agg_col] - predicted_aggr) / predicted_aggr
                    candidate_list[i].append([deviation, t_sim, row, predicted_aggr])
                    print(row, predicted_aggr, deviation, t_sim)
                    max_pred_aggr = max(max_pred_aggr, predicted_aggr)
                    min_pred_aggr = min(min_pred_aggr, predicted_aggr)
                    cnt += 1
                # print(len(candidate_list[i]), cnt)

                # print('In scoring: ', local_patterns[i])
                pred_aggr_range = max_pred_aggr - min_pred_aggr
                print(max_pred_aggr, min_pred_aggr)
                for k in range(cnt):
                    deviation = candidate_list[i][k][0]
                    t_sim = candidate_list[i][k][1]
                    predicted_aggr = candidate_list[i][k][3]
                    # denominator = agg_range * agg_range
                    if pred_aggr_range < 1e-8:
                        denominator = 1
                    else:
                        denominator = pred_aggr_range * pred_aggr_range
                    score = math.sqrt(t_sim * t_sim + deviation * deviation / denominator)
                    # score = abs(deviation)
                    # print(deviation, t_sim, score, candidate_list[i][k][2])
                    if deviation == 0:
                        top_k_lists[i].append((-1e10, i, candidate_list[i][k][2]))
                    else:
                        top_k_lists[i].append((-dir * deviation / abs(deviation) * score, i, candidate_list[i][k][2]))

                end = time.clock()
                score_computing_time += end - start
                    
        # uses heapq to manipulate merge of explanations from multiple constraints
        # start = time.clock()
        # merge_top_k_list = []
        # marked = {}
        # for i in range(len(cons)):
        #     heapify(top_k_lists[i])
        #     heappush(merge_top_k_list, heappop(top_k_lists[i]))

        # answer[j] = [{} for i in range(TOP_K)]
        # cnt = 0
        # while cnt < TOP_K:
        #     poped_tuple = heappop(merge_top_k_list)
        #     if poped_tuple[2] in marked:
        #         continue
        #     marked[poped_tuple[2]] = True
        #     answer[j][cnt] = (-poped_tuple[0], poped_tuple[1], poped_tuple[2])
        #     heappush(merge_top_k_list, heappop(top_k_lists[poped_tuple[1]]))
        #     cnt += 1
        # end = time.clock()
        # result_merging_time += end - start

        for i in range(len(local_patterns)):
            # print("TOP K: ", top_k_lists[i])
            if len(top_k_lists[i]) > 3:
                answer[j].append(sorted(top_k_lists[i], key=lambda x: x[0], reverse=True)[0:TOP_K])
            else:
                answer[j].append(top_k_lists[i])
            # print(len(answer[j][-1]))

        #answer[j] = top_k_lists


    print('Local pattern loading time: ' + str(local_pattern_loading_time) + 'seconds')
    print('Question validating time: ' + str(question_validating_time) + 'seconds')
    print('Score computing time: ' + str(score_computing_time) + 'seconds')
    #print('Result merging time: ' + str(result_merging_time) + 'seconds')
    return answer, local_patterns_list


def find_user_question(cur, global_patterns, table_name, max_uq_num, cat_sim):
    '''
        load user questions
    '''
    uq = []
    for pat in global_patterns:
        # print(pat)
        F = str(pat[0]).replace("\'", '')[1:-1]
        V = str(pat[1]).replace("\'", '')[1:-1]
        agg_query = '''SELECT {},{},{}({}) FROM {} GROUP BY {}, {};'''.format(
                F, V, pat[3], pat[2], table_name, F, V
            )
        cur.execute(agg_query)
        res = cur.fetchall()
        uq_cnt_cur_pat = 0
        for t in res:
            row_data = {}
            F_list = F.split(', ')
            V_list = V.split(', ')
            if 'month' not in V_list:
                break
            cnt = 0 
            for f_attr in F_list:
                if cat_sim.is_categorical(f_attr):
                    row_data[f_attr] = t[cnt]
                else:
                    if is_float(t[cnt]):
                        row_data[f_attr] = float(t[cnt])
                    elif is_integer(t[cnt]):
                        row_data[f_attr] = float(int(t[cnt]))
                    else:
                        row_data[f_attr] = t[cnt]
                cnt += 1
            for v_attr in V_list:
                if cat_sim.is_categorical(v_attr):
                    row_data[v_attr] = t[cnt]
                else:
                    if is_float(t[cnt]):
                        row_data[v_attr] = float(t[cnt])
                    elif is_integer(t[cnt]):
                        row_data[v_attr] = float(int(t[cnt]))
                    else:
                        row_data[v_attr] = t[cnt]
                cnt += 1
            row_data[pat[3] + '(' + pat[2] + ')'] = t[-1]
            l_pat = get_local_pattern(pat, row_data, cur, table_name)
            if isinstance(l_pat, dict):
                continue
            validate_res, predicted_aggr = validate_local_regression_pattern(
                l_pat, DEFAULT_EPSILON, row_data, 1, pat[3] + '(' + pat[2] + ')', 
                cur, table_name
            )
            if validate_res != 0:
                uq.append({'target_tuple': row_data, 'dir': validate_res})
                uq_cnt_cur_pat += 1
                if uq_cnt_cur_pat > 0:
                    break;
        if len(uq) >= max_uq_num:
            break
    print(uq)
    return uq[:max_uq_num]

def load_patterns(cur, pat_table_name):
    '''
        load pre-defined constraints(currently only fixed attributes and variable attributes)
    '''
    global_pattern_table = pat_table_name + '_global'
    load_query = "SELECT * FROM {};".format(global_pattern_table)
    cur.execute(load_query)
    res = cur.fetchall()
    print(res)
    patterns = []
    for pat in res:
        if pat[0].find('community_area') != -1 and pat[0].find('district') != -1:
            continue
        if pat[1].find('community_area') != -1 and pat[1].find('district') != -1:
            continue
        if pat[0].find('community_area') != -1 and pat[1].find('district') != -1:
            continue
        if pat[1].find('community_area') != -1 and pat[0].find('district') != -1:
            continue
        # if pat[1].find('community_area') == -1 and pat[1].find('district') == -1:
        #     continue
        # if pat[0].find('district') != -1 and pat[1].find('district') == -1:
        #     continue
        patterns.append(list(pat))
        patterns[-1][0] = patterns[-1][0][1:-1].replace(' ', '').split(',')
        patterns[-1][1] = patterns[-1][1][1:-1].replace(' ', '').split(',')
    
    return patterns

        
def output_explanation(outputfile, Q, explanations_list, local_patterns_list, global_patterns):
    ofile = sys.stdout
    if outputfile != '':
        ofile = open(outputfile, 'w')

    for i, top_k_list in enumerate(explanations_list):
        ofile.write('User question {}: {}\n'.format(str(i+1), str(Q[i])))
        print(len(top_k_list))
        for k, list_by_pat in enumerate(top_k_list):
            if list_by_pat[0] != 1:
                if k < len(global_patterns):
                    ofile.write('Globally held pattern: {}\n'.format(str(local_patterns_list[i][k])))
                else:
                    ofile.write('Locally held pattern: {}\n'.format(str(local_patterns_list[i][k])))
            print(len(list_by_pat))
            if isinstance(list_by_pat[0], int):
                if list_by_pat[0] == 1:
                    ofile.write("The local regression constraint derived from the " + str(i+1) +
                        "th constraint and the target tuple does not hold\n")
                elif list_by_pat[0] == -2:
                    ofile.write("The user question is invalid according to the local regression constraint derived from the "
                            + str(i+1) + "th constraint and the target tuple -- in the opposite direction\n")
                elif list_by_pat[0] == 2:
                    ofile.write("The user question is invalid according to the local regression constraint derived from the "
                            + str(i+1) + "th constraint and the target tuple -- the target tuple can be predicted by the constraint\n")
            else:

                if len(list_by_pat) > 2:
                    for j in range(TOP_K):
                        e = list_by_pat[j]
                        if e[0] < 0:
                            break
                        ofile.write('------------------------\n')
                        print_str = ''
                        e_tuple = list(e[2].values())
                        # e_tuple_str = ','.join(e_tuple.to_string(header=False,index=False,index_names=False).split('  ')[1:])
                        e_tuple_str = ','.join(map(str, e_tuple))
                        ofile.write('Top ' + str(j+1) + ' explanation:\n')
                        # ofile.write('Constraint ' + str(e[1]+1) + ': [' + ','.join(global_patterns[e[1]][0]) + ']' + '[' + ','.join(global_patterns[e[1]][1]) + ']')
                        ofile.write('Constraint ' + str(e[1]+1) + ': [' + ','.join(local_patterns_list[i][e[1]][0]) + ']' + 
                            '[' + ','.join(local_patterns_list[i][e[1]][1]) + ']' +
                            '[' + ','.join(local_patterns_list[i][e[1]][2]) + ']')
                        ofile.write('\n')
                        ofile.write('Score: ' + str(e[0]))
                        ofile.write('\n')
                        ofile.write('(' + e_tuple_str + ')')
                        ofile.write('\n')
                else:
                    ofile.write('------------------------\n')
                    ofile.write('Explanation:\n')
                    ofile.write(str(list_by_pat) + '\n')
            ofile.write('------------------------\n\n\n------------------------\n')
    ofile.close()

def main(argv=[]):
    query_result_table = DEFAULT_QUERY_RESULT_TABLE
    pattern_table = DEFAULT_PATTERN_TABLE
    user_question_number = DEFAULT_USER_QUESTION_NUMBER
    outputfile = ''
    epsilon = DEFAULT_EPSILON
    aggregate_column = DEFAULT_AGGREGATE_COLUMN
    try:
        # conn = psycopg2.connect("host=216.47.152.61 port=5432 dbname=postgres user=antiprov password=test")
        conn = psycopg2.connect("host=localhost port=5432 dbname=antiprov user=zjmiao password=keertijeff")
        cur = conn.cursor()
    except psycopg2.OperationalError:
        print('数据库连接失败！')

    try:
        opts, args = getopt.getopt(argv,"hq:p:un:o:e:a",["qtable=", "ptable=", "uqnum=","ofile=","epsilon=","aggregate_column="])
    except getopt.GetoptError:
        print('explanation.py -q <query_result_table> -p <pattern_table> -un <user_question_number> -o <outputfile> -e <epsilon>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print('explanation.py -q <query_result_table> -p <pattern_table> -un <user_question_number> -o <outputfile> -e <epsilon>')
            sys.exit(2)    
        elif opt in ("-q", "--qtable"):
            query_result_table = arg
        elif opt in ("-p", "--ptable"):
            pattern_table = arg
        elif opt in ("-un", "--uqnum"):
            user_question_number = int(arg)
        elif opt in ("-o", "--ofile"):
            outputfile = arg
        elif opt in ("-e", "--epsilon"):
            epsilon = float(arg)

    start = time.clock()
    category_similarity = CategorySimilarityNaive(cur, pattern_table, 
        [('block', 'block_location'), 
        ('community_area', 'community_area_location'),
        ('district', 'district_location')]
    )
    global_patterns = load_patterns(cur, pattern_table)
    print(global_patterns)
    Q = find_user_question(cur, global_patterns, pattern_table, user_question_number, category_similarity)
    
    # category_similarity = CategorySimilarityMatrix(EXAMPLE_SIMILARITY_MATRIX_PATH)
    #category_similarity = CategoryNetworkEmbedding(EXAMPLE_NETWORK_EMBEDDING_PATH, data['df'])
    #num_dis_norm = normalize_numerical_distance(data['df'])
    num_dis_norm = normalize_numerical_distance(cur=cur, table_name=query_result_table)
    end = time.clock()
    print('Loading time: ' + str(end-start) + 'seconds')
    
    start = time.clock()
    #regression_package = 'scikit-learn'
    regression_package = 'statsmodels'
    explanations_list, local_patterns_list = find_explanation_regression_based(Q, global_patterns, category_similarity, 
                                                          num_dis_norm, epsilon, 
                                                          cur, pattern_table, query_result_table)
    end = time.clock()
    print('Total querying time: ' + str(end-start) + 'seconds')

    

    # for i, explanations in enumerate(explanations_list):
    #     ofile.write('User question ' + str(i+1) + ':\n')
    #     for j, e in enumerate(explanations):
    #         print_str = ''
    #         e_tuple = data['df'].loc[data['df']['index'] == e[2]]
    #         e_tuple_str = ','.join(e_tuple.to_string(header=False,index=False,index_names=False).split('  ')[1:])
    #         ofile.write('Top ' + str(j+1) + ' explanation:\n')
    #         ofile.write('Constraint ' + str(e[1]+1) + ': [' + ','.join(constraints[e[1]][0]) + ']' + '[' + ','.join(constraints[e[1]][1]) + ']')
    #         ofile.write('\n')
    #         ofile.write('Score: ' + str(e[0]))
    #         ofile.write('\n')
    #         ofile.write('(' + e_tuple_str + ')')
    #         ofile.write('\n')
    #     ofile.write('------------------------\n')

    output_explanation(outputfile, Q, explanations_list, local_patterns_list, global_patterns)
    print(Q)
if __name__ == "__main__":
    main(sys.argv[1:])





"""
Module wrapping database access.
"""
import sys
import logging
import sqlalchemy as sa
import psycopg2

log = logging.getLogger(__name__)

# ********************************************************************************
class AttrDef:

    name=None
    datatype='varchar'

    def __init__(self, name, datatype='varchar'):
        self.name=name
        self.datatype = datatype

    def toSQL(self):
        return "{} {}".format(name, datatype)

# ********************************************************************************
class DBConnection:
    """
    DBConnection wraps an sqlalchemy database connection (currently only postgres through psycopg2).
    """

    def __init__ (self,
                  host='127.0.0.1',
                  user='postgres',
                  port=5432,
                  db='postgres',
                  password=None):
        self.host=host
        self.user=user
        self.port=port
        self.db=db
        self.password=password
        self.engine = None
        self.conn=None
        log.debug("create DBconnection info: %s", self.__dict__)


    # overwrite __getitem__ to allow dictory style access to options
    def __getitem__(self, key):
        if key not in self.__dict__:
            raise AttributeError("No such attribute: " + key)
        return self.__dict__[key]

    def __setitem__(self,key,value):
        if key not in self.__dict__:
            raise AttributeError("No such attribute: " + key)
        self.__dict__[key] = value
    
    def getValidKeys(self):
        return self.__dict__
        
    def getUrl(self):
        url = 'postgresql://{}:{}@{}:{:d}/{}'.format(self.user,
                                                  ('' if self.password is None else self.password),
                                                  self.host,
                                                  self.port,
                                                  self.db)
        log.debug("connection url %s", url)
        return url

    def connect(self):
        try:
            engine = sa.create_engine(self.getUrl(),
                echo=False)
            conn = engine.connect()
            log.debug("connected to DB")
        except Exception as ex:
            print(type(ex))
            print("\n\n",ex.args)
            sys.exit(1)
        return conn

    def getPostgresConnectStr(self):
        connstr = 'host={} port={} dbname={} user={}'.format(self.host,
                                                          self.port,
                                                          self.db,
                                                          self.user)
        connstr += ('' if self.password is None else ' password={}'.format(self.password))
        log.debug("psycopg2 connection string is <%s>", connstr)
        return connstr
    
    def pgconnect(self):
        try:
             self.conn = psycopg2.connect(self.getPostgresConnectStr())
             return self.conn
        except psycopg2.OperationalError as ex:
             log.error('Fail to connect to the database ({})\n\n{}'.format(self.getUrl(), ex.args))
             sys.exit(1)

    
    def close(self):
        if conn is not None:
            conn.close()
            conn = None
        if engine is not None:
            engine.dispose()
        engine=None
        log.debug("closed connection")

    CREATE_TABLE_TEMP = 'CREATE TABLE IF NOT EXISTS {} ({});'

    def createOrReplaceTable(tableName, attrs, constraints):
        if engine is None or conn is None:
            raise SQLException("cannot create table: no database connection")
        stmtStr = CREATE_TABLE_TEMP.format(tableName, ",".join(attrs + constraints))
        conn.execute(stmtStr)


if __name__ =='__main__':
    dbc = DBConnection()
    dbc['local_table'] = 'pub_large_no_domain_local'
    print(dbc['local_table'])



#!/usr/bin/python
# -*- coding:utf-8 -*- 

import sys, getopt
import pandas
import csv
#import statsmodels.formula.api as smf
from sklearn import preprocessing
import math
import time
from heapq import *
import logging
from capexplain.similarity.category_similarity_matrix import *
from capexplain.similarity.category_network_embedding import *
from capexplain.utils import *
from capexplain.pattern_model.LocalRegressionPattern import *
from capexplain.cl.cfgoption import DictLike

# setup logging
log = logging.getLogger(__name__)

#********************************************************************************
# Configuration for Explanation generation
class ExplConfig(DictLike):

    DEFAULT_RESULT_PATH = './input/query_res.csv'
    DEFAULT_QUESTION_PATH = './input/user_question.csv'
    DEFAULT_CONSTRAINT_PATH = './input/CONSTRAINTS'
    EXAMPLE_NETWORK_EMBEDDING_PATH = './input/NETWORK_EMBEDDING'
    EXAMPLE_SIMILARITY_MATRIX_PATH = './input/SIMILARITY_DEFINITION'
    DEFAULT_AGGREGATE_COLUMN = 'count'
    DEFAULT_CONSTRAINT_EPSILON = 0.05
    TOP_K = 5
    REGRESSION_PACKAGES = [ 'scikit-learn', 'statsmodels' ]
    
    def __init__(self,
                 query_result_file = DEFAULT_RESULT_PATH,
                 constraint_file = DEFAULT_CONSTRAINT_PATH,
                 user_question_file = DEFAULT_QUESTION_PATH,
                 outputfile = '',
                 constraint_epsilon = DEFAULT_CONSTRAINT_EPSILON,
                 aggregate_column = DEFAULT_AGGREGATE_COLUMN,
                 regression_package = 'statsmodels'
    ):
        self.query_result_file = query_result_file
        self.constraint_file = constraint_file
        self.user_question_file = user_question_file
        self.outputfile = outputfile
        self.constraint_epsilon = constraint_epsilon
        self.aggregate_column = aggregate_column
        self.regression_package = regression_package

    def __str__(self):
        return self.__dict__.__str__()

#********************************************************************************
# construct local patterns
def build_local_regression_constraint(data, column_index, t, con, epsilon, agg_col, regression_package):
    """Build local regression constraint from Q(R), t, and global regression constraint

    Args:
        data: result of Q(R)
        column_index: index for values in each column
        t: target tuple in Q(R)
        con: con[0] is the list of fixed attributes in Q(R), con[1] is the list of variable attributes in Q(R)
        epsilon: threshold for local regression constraint
        regression_package: which package is used to compute regression 
    Returns:
        A LocalRegressionPattern object whose model is trained on \pi_{con[1]}(Q_{t[con[0]]}(R))
    """
    tF = get_F_value(con[0], t)
    local_con = LocalRegressionPattern(con[0], tF, con[1], agg_col, epsilon)
    train_data = {agg_col: []}
    for v in con[1]:
        train_data[v] = []
    # for index, row in data['df'].iterrows():
    #     if get_F_value(con[0], row) == tF:
    #         for v in con[1]:
    #             train_data[v].append(row[v])
    #         train_data[agg_col].append(row[agg_col])

    
    for idx in column_index[con[0][0]][tF[0]]:
        row = data['df'].loc[data['df']['index'] == idx]
        row = row.to_dict('records')[0]
        #print row
        if get_F_value(con[0], row) == tF:
            for v in con[1]:
                train_data[v].append(row[v])
            train_data[agg_col].append(row[agg_col])
    if regression_package == 'scikit-learn':
        train_x = {}
        for v in con[1]:
            if v in data['le']:
                train_data[v] = data['le'][v].transform(train_data[v])
                train_data[v] = data['ohe'][v].transform(train_data[v].reshape(-1, 1))
                #print data['ohe'][v].transform(train_data[v].reshape(-1, 1))
                train_x[v] = train_data[v]
            else:
                if v != agg_col:
                    train_x[v] = np.array(train_data[v]).reshape(-1, 1)
        train_y = np.array(train_data[agg_col]).reshape(-1, 1)
        train_x = np.concatenate(list(train_x.values()), axis=-1)
        local_con.train_sklearn(train_x, train_y)
    else:
        #train_data = pandas.DataFrame(train_data)
        formula = agg_col + ' ~ ' + ' + '.join(con[1])
        print 
        local_con.train(train_data, formula)
    return local_con

def validate_local_regression_constraint(data, local_con, t, dir, agg_col, regression_package):
    """Check the validicity of the user question under a local regression constraint

    Args:
        data: data['df'] is the data frame storing Q(R)
            data['le'] is the label encoder, data['ohe'] is the one-hot encoder
        local_con: a LocalRegressionPattern object
        t: target tuple in Q(R)
        dir: whether user thinks t[agg(B)] is high or low
        agg_col: the column of aggregated value
        regression_package: which package is used to compute regression 
    Returns:
        the actual direction that t[agg(B)] compares to its expected value, and the expected value from local_con
    """
    test_tuple = {}
    for v in local_con.var_attr:
        test_tuple[v] = [t[v]]
    if regression_package == 'scikit-learn':
        for v in local_con.var_attr:
            if v in data['le']:
                test_tuple[v] = data['le'][v].transform(test_tuple[v])
                test_tuple[v] = data['ohe'][v].transform(test_tuple[v].reshape(-1, 1))
            else:
                test_tuple[v] = np.array(test_tuple[v]).reshape(-1, 1)
        
        test_tuple = np.concatenate(list(test_tuple.values()), axis=-1)
        predictY = local_con.predict_sklearn(test_tuple)
    else:
        predictY = local_con.predict(pandas.DataFrame(test_tuple))

    if t[agg_col] < (1-local_con.epsilon) * predictY[0]:
        return -dir, predictY[0]
    elif t[agg_col] > (1+local_con.epsilon) * predictY[0]:
        return dir, predictY[0]
    else:
        return 0, predictY[0]
        
def tuple_similarity(t1, t2, var_attr, cat_sim, num_dis_norm, agg_col):
    """Compute the similarity between two tuples t1 and t2 on their attributes var_attr

    Args:
        t1, t2: two tuples
        var_attr: variable attributes
        cat_sim: the similarity measure for categorical attributes
        num_dis_norm: normalization terms for numerical attributes
        agg_col: the column of aggregated value
    Returns:
        the Gower similarity between t1 and t2
    """
    sim = 0.0
    cnt = 0
    for col in var_attr:
        if t1[col] is None or t2[col] is None:
            continue
        if cat_sim.is_categorical(col):
            s = cat_sim.compute_similarity(col, t1[col], t2[col], agg_col)
            sim += s
        else:
            if col != agg_col and col != 'index':
                temp = abs(t1[col] - t2[col]) / num_dis_norm[col]['range']
                sim += 1-temp
        cnt += 1
    return sim / cnt

def find_explanation_regression_based(data, user_question_list, cons, cat_sim, num_dis_norm, cons_epsilon, agg_col, regression_package):

    """Find explanations for user questions

    Args:
        data: data['df'] is the data frame storing Q(R)
            data['le'] is the label encoder, data['ohe'] is the one-hot encoder
        user_question_list: list of user questions (t, dir), all questions have the same Q(R)
        cons: list of fixed attributes and variable attributes of global constraints
        cat_sim: the similarity measure for categorical attributes
        num_dis_norm: normalization terms for numerical attributes
        cons_epsilon: threshold for local regression constraints
        agg_col: the column of aggregated value
        regression_package: which package is used to compute regression 
    Returns:
        the top-k list of explanations for each user question
    """

    answer = [[] for i in range(len(user_question_list))]
    index_building_time = 0
    constraint_building_time = 0
    question_validating_time = 0
    score_computing_time = 0
    result_merging_time = 0

    start = time.clock()
    column_index = dict()
    for column in data['df']:
        column_index[column] = dict()
    for index, row in data['df'].iterrows():
        for column in data['df']:
            val = row[column]
            if not val in column_index[column]:
                column_index[column][val] = []
            column_index[column][val].append(index)
    end = time.clock()
    index_building_time += end - start

    for j, uq in enumerate(user_question_list):
        dir = uq['dir']
        t = uq['target_tuple']
        print(uq)
        candidate_list = [[[] for i in range(len(data['df'].index))] for i in range(len(cons))]
        top_k_lists = [[(1e10,0,0) for i in range(len(data['df'].index))] for i in range(len(cons))]
        validate_res_list = []

        psi = []
        local_cons = []
        start = time.clock()
        for i in range(len(cons)):
            local_cons.append(build_local_regression_constraint(data, column_index, t, cons[i], cons_epsilon, agg_col, regression_package))
            local_cons[i].print_fit_summary()
        end = time.clock()
        constraint_building_time += end - start

        explanation_type = 0
        for i in range(0, len(local_cons)):
            psi.append(0)
            start = time.clock()
            validate_res, predicted_aggr = validate_local_regression_constraint(data, local_cons[i], t, dir, agg_col, regression_package)
            validate_res_list.append(validate_res)
            end = time.clock()
            question_validating_time += end - start

            if validate_res < -1:
                print ("The local regression constraint derived from the " + str(i+1) +
                        "th constraint and the target tuple does not hold")
                explanation_type = 1
            elif validate_res == -1:
                print ("The user question is invalid according to the local regression constraint derived from the "
                        + str(i+1) + "th constraint and the target tuple -- in the opposite direction")
                explanation_type = 2
            elif validate_res == 0:
                print ("The user question is invalid according to the local regression constraint derived from the "
                        + str(i+1) + "th constraint and the target tuple -- "
                        "the target tuple can be predicted by the constraint")
                explanation_type = 2
            else:
                print ("The user question is valid according to the local regression constraint derived from the "
                        + str(i+1) + "th constraint and the target tuple")
                explanation_type = 3
                #deviation = dir * (t[agg_col] - predicted_aggr) / predicted_aggr
                cnt = 0
                #for index, row in data['df'].iterrows():
                start = time.clock()
                f_value = get_F_value(local_cons[i].fix_attr, t)
                for idx in column_index[local_cons[i].fix_attr[0]][f_value[0]]:
                    row = data['df'].loc[data['df']['index'] == idx]
                    row = row.to_dict('records')[0]
                    if not local_cons[i].same_fixed_attributes(row):
                        continue
                    psi[-1] += row[agg_col]
                    t_sim = tuple_similarity(t, row, local_cons[i].var_attr, cat_sim, num_dis_norm, agg_col)
                    # print local_cons[i].var_attr, row
                    test_tuple = dict(zip(local_cons[i].var_attr, 
                                          map(lambda x:[x], get_V_value(local_cons[i].var_attr, row))))
                    predicted_aggr = []
                    if regression_package == 'scikit-learn':
                        for v in local_cons[i].var_attr:
                            if v in data['le']:
                                test_tuple[v] = data['le'][v].transform(test_tuple[v])
                                test_tuple[v] = data['ohe'][v].transform(test_tuple[v].reshape(-1, 1))
                            else:
                                test_tuple[v] = np.array(test_tuple[v]).reshape(-1, 1)
                        predicted_aggr = local_cons[i].predict_sklearn(np.concatenate(list(test_tuple.values()), axis=-1))
                    else:
                        predicted_aggr = local_cons[i].predict(pandas.DataFrame(test_tuple)).tolist()
                    deviation = (row[agg_col] - predicted_aggr[0]) / predicted_aggr[0]
                    #print get_F_value(local_cons[i].fix_attr, row), get_V_value(local_cons[i].var_attr, row), predicted_aggr[0], deviation, t_sim
                    candidate_list[i][cnt] = [deviation, t_sim, row['index'], predicted_aggr[0]]
                    print(row, predicted_aggr[0])
                    cnt += 1
                
                for k in range(cnt):
                    deviation = candidate_list[i][k][0]
                    t_sim = candidate_list[i][k][1]
                    predicted_aggr = candidate_list[i][k][3]
                    denominator = 10
                    #denominator = psi[-1] * psi[-1]
                    #denominator = psi[-1] * psi[-1] / (predicted_aggr * predicted_aggr)
                    score = math.sqrt(t_sim * t_sim + deviation * deviation / denominator)
                    #print score
                    top_k_lists[i][k] = (dir * deviation / abs(deviation) * score, i, candidate_list[i][k][2])

                end = time.clock()
                score_computing_time += end - start
                    
        # uses heapq to manipulate merge of explanations from multiple constraints
        # start = time.clock()
        # merge_top_k_list = []
        # marked = {}
        # for i in range(len(cons)):
        #     heapify(top_k_lists[i])
        #     heappush(merge_top_k_list, heappop(top_k_lists[i]))

        # answer[j] = [{} for i in range(TOP_K)]
        # cnt = 0
        # while cnt < TOP_K:
        #     poped_tuple = heappop(merge_top_k_list)
        #     if poped_tuple[2] in marked:
        #         continue
        #     marked[poped_tuple[2]] = True
        #     answer[j][cnt] = (-poped_tuple[0], poped_tuple[1], poped_tuple[2])
        #     heappush(merge_top_k_list, heappop(top_k_lists[poped_tuple[1]]))
        #     cnt += 1
        # end = time.clock()
        # result_merging_time += end - start
        for i in range(len(local_cons)):
            answer[j].append(sorted(top_k_lists[i], key=lambda x: x[0]))

        #answer[j] = top_k_lists


    print('Index building time: ' + str(index_building_time) + 'seconds')
    print('Constraint building time: ' + str(constraint_building_time) + 'seconds')
    print('Question validating time: ' + str(question_validating_time) + 'seconds')
    print('Score computing time: ' + str(score_computing_time) + 'seconds')
    #print('Result merging time: ' + str(result_merging_time) + 'seconds')
    return answer

def load_data(qr_file=ExplConfig.DEFAULT_RESULT_PATH):
    ''' 
        load query result
    '''
    df = pandas.read_csv(open(qr_file, 'r'), header=0, quotechar="'")
    le = {}
    ohe = {}
    for column in df:
        df[column] = df[column].apply(lambda x: x.replace('\'', '').strip())
        df[column] = df[column].apply(lambda x: float_or_integer(x))
        # if it is a categorical attribute, first encode each one into integers, and then use one-hot encoding
        if df[column].dtype.kind == 'S' or df[column].dtype.kind == 'O':
            le[column] = preprocessing.LabelEncoder()
            le[column].fit(df[column])
            ohe[column] = preprocessing.OneHotEncoder()
            le_col = le[column].transform(df[column])
            le_col = le_col.reshape(-1, 1)
            ohe[column] = preprocessing.OneHotEncoder(sparse=False)
            ohe[column].fit(le_col)
    df.insert(0, 'index', range(0, len(df))) 
    data = {'df':df, 'le':le, 'ohe':ohe}
    return data

def load_user_question(uq_path=ExplConfig.DEFAULT_QUESTION_PATH):
    '''
        load user questions
    '''
    uq = []
    with open(uq_path, 'rt') as uqfile:
        reader = csv.DictReader(uqfile, quotechar='\'')
        headers = reader.fieldnames
        #temp_data = csv.reader(uqfile, delimiter=',', quotechar='\'')
        #for row in temp_data:
        for row in reader:
            row_data = {}
            for k, v in enumerate(headers):
                print(k, v)
                if v != 'direction':
                    if is_float(row[v]):
                        row_data[v] = float(row[v])
                    elif is_integer(row[v]):
                        row_data[v] = float(long(row[v]))
                    else:
                        row_data[v] = row[v]
            if row['direction'][0] == 'h':
                dir = 1
            else:
                dir = -1
            uq.append({'target_tuple': row_data, 'dir':dir})
    return uq

def load_constraints(cons_path=ExplConfig.DEFAULT_CONSTRAINT_PATH):
    '''
        load pre-defined constraints(currently only fixed attributes and variable attributes)
    '''
    inf = open(cons_path, 'r')
    cons = []
    while True:
        line = inf.readline()
        if not line:
            break
        cons.append([[],[]])
        cons[-1][0] = line.strip(' \r\n').split(',')
        line = inf.readline()
        cons[-1][1] = line.strip(' \r\n').split(',')
    inf.close()
    return cons

class ExplanationGenerator:

    def __init__(self, config : ExplConfig):
        self.config = config

    def doExplain(self):
        c=self.config
        log.info("start explaining ...")
        start = time.clock()
        data = load_data(c.query_result_file)
        log.debug("loaded query results from file")
        constraints = load_constraints(ExplConfig.DEFAULT_CONSTRAINT_PATH)
        log.debug("loaded patterns from file")
        Q = load_user_question(c.user_question_file)
        log.debug("loaded user question from file")
        category_similarity = CategorySimilarityMatrix(ExplConfig.EXAMPLE_SIMILARITY_MATRIX_PATH)
        #category_similarity = CategoryNetworkEmbedding(EXAMPLE_NETWORK_EMBEDDING_PATH, data['df'])
        num_dis_norm = normalize_numerical_distance(data['df'])
        end = time.clock()
        log.debug("done loading")
        print('Loading time: ' + str(end-start) + 'seconds')

        log.debug("start finding explanations ...")
        start = time.clock()
        #regression_package = 'scikit-learn'
        regression_package = 'statsmodels'
        explanations_list = find_explanation_regression_based(data, Q, constraints, category_similarity, 
                                                              num_dis_norm, c.constraint_epsilon, 
                                                              aggregate_column, regression_package)
        end = time.clock()
        print('Total querying time: ' + str(end-start) + 'seconds')
        log.debug("finding explanations ... DONE")
        
        ofile = sys.stdout
        if outputfile != '':
            ofile = open(outputfile, 'w')

        for i, top_k_list in enumerate(explanations_list):
            ofile.write('User question ' + str(i+1) + ':\n')
            for k, list_by_con in enumerate(top_k_list):
                for j in range(5):
                    e = list_by_con[j]
                    ofile.write('------------------------\n')
                    print_str = ''
                    e_tuple = data['df'].loc[data['df']['index'] == e[2]]
                    e_tuple_str = ','.join(e_tuple.to_string(header=False,index=False,index_names=False).split('  ')[1:])
                    ofile.write('Top ' + str(j+1) + ' explanation:\n')
                    ofile.write('Constraint ' + str(e[1]+1) + ': [' + ','.join(constraints[e[1]][0]) + ']' + '[' + ','.join(constraints[e[1]][1]) + ']')
                    ofile.write('\n')
                    #ofile.write('Score: ' + str(e[0]))
                    ofile.write('Score: ' + str(-e[0]))
                    ofile.write('\n')
                    ofile.write('(' + e_tuple_str + ')')
                    ofile.write('\n')
            
                ofile.write('------------------------\n')
        
        
def main(argv=[]):
    query_result_file = ExplConfig.DEFAULT_RESULT_PATH
    constraint_file = ExplConfig.DEFAULT_CONSTRAINT_PATH
    user_question_file = ExplConfig.DEFAULT_QUESTION_PATH
    outputfile = ''
    constraint_epsilon = ExplConfig.DEFAULT_CONSTRAINT_EPSILON
    aggregate_column = ExplConfig.DEFAULT_AGGREGATE_COLUMN
    try:
        opts, args = getopt.getopt(argv,"hq:c:u:o:e:a",["help","qfile=","cfile=","ufile=","ofile=","epsilon=","aggregate_column="])
    except getopt.GetoptError:
        print('explanation.py -q <query_result_file> -c <constraint_file> -u \
        <user_question_file> -o <outputfile> -e <epsilon> -a <aggregate_column>')
        sys.exit(2)
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            print('explanation.py -q <query_result_file> -c <constraint_file> -u \
            <user_question_file> -o <outputfile> -e <epsilon> -a <aggregate_column>')
            sys.exit(2)    
        elif opt in ("-q", "--qfile"):
            query_result_file = arg
        elif opt in ("-c", "--cfile"):
            constraint_file = arg
        elif opt in ("-u", "--ufile"):
            user_question_file = arg
        elif opt in ("-o", "--ofile"):
            outputfile = arg
        elif opt in ("-e", "--epsilon"):
            constraint_epsilon = float(arg)
        elif opt in ("-a", "--aggcolumn"):
            aggregate_column = arg

    print(opts)
    start = time.clock()
    data = load_data(query_result_file)
    constraints = load_constraints(ExplConfig.DEFAULT_CONSTRAINT_PATH)
    Q = load_user_question(user_question_file)
    category_similarity = CategorySimilarityMatrix(ExplConfig.EXAMPLE_SIMILARITY_MATRIX_PATH)
    #category_similarity = CategoryNetworkEmbedding(ExplConfig.EXAMPLE_NETWORK_EMBEDDING_PATH, data['df'])
    num_dis_norm = normalize_numerical_distance(data['df'])
    end = time.clock()
    print('Loading time: ' + str(end-start) + 'seconds')
    
    start = time.clock()
    #regression_package = 'scikit-learn'
    regression_package = 'statsmodels'
    explanations_list = find_explanation_regression_based(data, Q, constraints, category_similarity, 
                                                          num_dis_norm, constraint_epsilon, 
                                                          aggregate_column, regression_package)
    end = time.clock()
    print('Total querying time: ' + str(end-start) + 'seconds')

    ofile = sys.stdout
    if outputfile != '':
        ofile = open(outputfile, 'w')

    # for i, explanations in enumerate(explanations_list):
    #     ofile.write('User question ' + str(i+1) + ':\n')
    #     for j, e in enumerate(explanations):
    #         print_str = ''
    #         e_tuple = data['df'].loc[data['df']['index'] == e[2]]
    #         e_tuple_str = ','.join(e_tuple.to_string(header=False,index=False,index_names=False).split('  ')[1:])
    #         ofile.write('Top ' + str(j+1) + ' explanation:\n')
    #         ofile.write('Constraint ' + str(e[1]+1) + ': [' + ','.join(constraints[e[1]][0]) + ']' + '[' + ','.join(constraints[e[1]][1]) + ']')
    #         ofile.write('\n')
    #         ofile.write('Score: ' + str(e[0]))
    #         ofile.write('\n')
    #         ofile.write('(' + e_tuple_str + ')')
    #         ofile.write('\n')
    #     ofile.write('------------------------\n')

    for i, top_k_list in enumerate(explanations_list):
        ofile.write('User question ' + str(i+1) + ':\n')
        for k, list_by_con in enumerate(top_k_list):
            for j in range(5):
                e = list_by_con[j]
                ofile.write('------------------------\n')
                print_str = ''
                e_tuple = data['df'].loc[data['df']['index'] == e[2]]
                e_tuple_str = ','.join(e_tuple.to_string(header=False,index=False,index_names=False).split('  ')[1:])
                ofile.write('Top ' + str(j+1) + ' explanation:\n')
                ofile.write('Constraint ' + str(e[1]+1) + ': [' + ','.join(constraints[e[1]][0]) + ']' + '[' + ','.join(constraints[e[1]][1]) + ']')
                ofile.write('\n')
                #ofile.write('Score: ' + str(e[0]))
                ofile.write('Score: ' + str(-e[0]))
                ofile.write('\n')
                ofile.write('(' + e_tuple_str + ')')
                ofile.write('\n')
            
            ofile.write('------------------------\n')

if __name__ == "__main__":
    main(sys.argv[1:])





import tkinter as tk
from tkinter import *
from tkinter import ttk
import tkinter.messagebox
from tkinter import filedialog
from tkinter import font
from tkinter.font import Font,nametofont
import pandas as pd
import psycopg2
from pandastable import TableModel
from pandastable import PlotViewer
from pandastable import Table
import re
from capexplain.explain.explanation import ExplanationGenerator
from capexplain.explain.explanation import ExplConfig
import matplotlib
matplotlib.use("TkAgg")
from matplotlib.backends.backend_tkagg import (
	FigureCanvasTkAgg, NavigationToolbar2Tk)
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import numpy as np
import logging
from mpl_toolkits.mplot3d import Axes3D
from capexplain.gui.User_Query_Frame import User_Query_Frame
from capexplain.gui.DBinfo import DBinfo
from capexplain.gui.Pattern_Frame import Local_Pattern_Frame
from capexplain.gui.Exp_Frame import Exp_Frame
from capexplain.cl.cfgoption import DictLike
from capexplain.database.dbaccess import DBConnection



#to do
# 1. fix "when misinput system have to be restart" issue
# 2. fix "drill down" in crime data set

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s %(levelname)s %(name)s line %(lineno)d: %(message)s')
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)


# conn = psycopg2.connect(dbname="antiprov",user="antiprov",host="127.0.0.1",port="5436")
class CAPE_UI:

	def __init__(self,parent,conn,config):
		self.conn=conn
		self.config=config
		self.cur=self.conn.cursor()
		# self.assigned_local_table = assigned_local_table
		# self.assigned_global_table = assigned_global_table

#----------------------------main frame----------------------------------------#

		self.parent=parent
		self.main_frame=ttk.Frame(self.parent,padding=(3,3,12,12))

		self.main_frame.columnconfigure(0,weight=1)
		self.main_frame.columnconfigure(1,weight=8,uniform=1)
		self.main_frame.columnconfigure(2,weight=7,uniform=1)

		self.main_frame.rowconfigure(0,weight=3)
		self.main_frame.rowconfigure(1,weight=2)
		self.main_frame.rowconfigure(2,weight=2)
		self.main_frame.rowconfigure(3,weight=2)
		self.main_frame.rowconfigure(4,weight=2)
		self.main_frame.rowconfigure(5,weight=2)
		self.main_frame.rowconfigure(6,weight=2)
		self.main_frame.rowconfigure(7,weight=2)
		self.main_frame.rowconfigure(8,weight=2)
		self.main_frame.rowconfigure(9,weight=2)

#----------------------------------place frames----------------------------------#
		
		self.main_frame.grid(column=0, row=0, columnspan=3, rowspan=10, sticky='nsew')

		self.table_frame = ttk.Frame(self.main_frame, borderwidth=5, relief="sunken")
		self.table_frame.grid(column=0,row=0, columnspan=1, rowspan=10, sticky='nsew')

		self.query_frame = ttk.Frame(self.main_frame, borderwidth=5, relief="ridge")
		self.query_frame.grid(column=1, row=0, columnspan=1, rowspan=2, sticky='nsew')

		self.query_frame.rowconfigure(0,weight=5)
		self.query_frame.rowconfigure(0,weight=1)

		self.query_template_frame = ttk.Frame(self.query_frame)
		self.query_template_frame.grid(row=0,sticky='nsew')

		self.query_button_frame = ttk.Frame(self.query_frame)
		self.query_button_frame.grid(row=1,sticky='nsew')


		self.query_result = ttk.Frame(self.main_frame, borderwidth=5, relief="sunken")
		self.query_result.grid(column=1, row=2, columnspan=1, rowspan=8, sticky='nsew')

		self.local_pattern = ttk.Frame(self.main_frame, borderwidth=5, relief="sunken")
		self.local_pattern.grid(column=2, row=0, columnspan=2, rowspan=5, sticky='nsew')

		self.explanation_frame = ttk.Frame(self.main_frame, borderwidth=5, relief="sunken")
		self.explanation_frame.grid(column=2, row=5, columnspan=2, rowspan=5, sticky='nsew')


#---------------------------table frame-----------------------------------------#
		self.table_view = ttk.Treeview(self.table_frame,height=47)
		self.table_info = ttk.Label(self.table_frame, text="Database Information",font=('Times New Roman bold',11),borderwidth=5,relief=RIDGE)
		self.table_info.grid(column=0, row=0,sticky='nsew')
		self.table_view.grid(column=0, row=1)

		self.db_info = DBinfo(conn=self.conn)
		self.db_info_dict = self.db_info.get_db_info()

		table_index = 0
		for key,value in self.db_info_dict.items(): 
		   self.table_view.insert('', 'end','item'+str(table_index),text = key)
		   for n in value:
			   self.table_view.insert('item'+str(table_index),'end',text=n)
		   table_index +=1

		# self.pub_dict = {"dict_name":"pub",
		# "global_name":"dev.pub_global",
		# "local_name":"dev.pub_local"
		# }

		# self.crime_dict = {"dict_name":"crime",
		# "global_name": "dev.crime_global",
		# "local_name":"dev.crime_local"
		# }


#----------------------------Query frame----------------------------------------#

		self.query_temp = User_Query_Frame(conn=self.conn,table_dict=self.db_info_dict,parent=self.query_frame)

		self.query_button = Button(self.query_button_frame,text="Run Query",font=('Times New Roman bold',12),command=self.run_query)
		self.query_button.pack(side=RIGHT)
		self.show_global_pattern_button = Button(self.query_button_frame,text="Show Global Pattern",font=('Times New Roman bold',12),command=self.show_global_pattern)
		self.show_global_pattern_button.pack(side=RIGHT)
		self.show_local_pattern_button = Button(self.query_button_frame,text="Show Local Pattern",font=('Times New Roman bold',12),command=self.show_local_pattern)
		self.show_local_pattern_button.pack(side=RIGHT)

#----------------------------Query result frame --------------------------------#
		self.query_result.columnconfigure(0,weight=6)
		self.query_result.columnconfigure(1,weight=1)
		self.query_result.rowconfigure(0,weight=1)
		self.query_result.rowconfigure(1,weight=10)
		self.query_result.rowconfigure(2,weight=10)

		self.result_label = Label(self.query_result,text='Query Result',font=('Times New Roman bold',12),borderwidth=5,relief=RIDGE)
		self.result_label.grid(row=0,column=0,sticky='nsew')

		self.high_low_frame = Frame(self.query_result)
		self.high_low_frame.grid(column=1,row=1,rowspan=2,sticky='nsew')
		
		self.high_low_frame.columnconfigure(0,weight=1)
		self.high_low_frame.rowconfigure(0,weight=1)
		self.high_low_frame.rowconfigure(1,weight=1)
		self.high_low_frame.rowconfigure(2,weight=1)
		self.high_low_frame.rowconfigure(3,weight=1)
		self.high_low_frame.rowconfigure(4,weight=1)
		self.high_low_frame.rowconfigure(5,weight=1)
		self.high_low_frame.rowconfigure(6,weight=1)
		self.high_low_frame.rowconfigure(7,weight=1)


		self.high_button = Button(self.high_low_frame,text='High',font=('Times New Roman bold',12), command=self.handle_high)
		self.high_button.grid(column=0,row=1)

		self.low_button = Button(self.high_low_frame,text='Low',font=('Times New Roman bold',12), command=self.handle_low)
		self.low_button.grid(column=0,row=2)

		self.show_results = Frame(self.query_result)
		self.show_results.grid(column=0,row=1,sticky='nsew')

		self.show_global = Frame(self.query_result)
		self.show_global.grid(column=0,row=2,sticky='nsew')

		self.query_result_table = Table(self.show_results)
		self.query_result_table.show()

		self.parent.columnconfigure(0, weight=1)
		self.parent.rowconfigure(0, weight=1)

#---------------------------Global Pattern Frame -----------------------------------#
		
		self.show_global.rowconfigure(0,weight=1)
		self.show_global.rowconfigure(1,weight=10)
		self.show_global.columnconfigure(0,weight=1)

		self.global_pattern_label = Label(self.show_global,text="Global Patterns",font=('Times New Roman bold',12),borderwidth=5,relief=RIDGE)
		self.global_pattern_label.grid(column=0,row=0,sticky='nsew')

		self.show_global_patterns = Frame(self.show_global)
		self.show_global_patterns.grid(column=0,row=1,sticky='nsew')

		self.global_description_button = Button(self.high_low_frame,text='Describe\nGlobal',font=('Times New Roman bold',12),command=self.global_description)
		self.global_description_button.grid(column=0,row=5)

		self.global_pattern_filter_button = Button(self.high_low_frame,text='Filter \nLocal\n Pattern',font=('Times New Roman bold',12),command=self.use_global_filter_local)
		self.global_pattern_filter_button.grid(column=0,row=6)

		self.global_pattern_table = Table(self.show_global_patterns)
		self.global_pattern_table.show()

		# create a sql function for sorting the array in order to be used for filtering

		sort_array_function = "CREATE OR REPLACE FUNCTION array_sort(anyarray) RETURNS anyarray AS $$"+\
		"SELECT array_agg(x order by x) FROM unnest($1) x;"+\
		"$$ LANGUAGE sql;"

		# logger.debug(sort_array_function)
		self.cur.execute(sort_array_function)	

#------------------------------- local pattern frame---------------------------------------#
		self.local_pattern.rowconfigure(0,weight=1)
		self.local_pattern.rowconfigure(1,weight=20)
		self.local_pattern.rowconfigure(2,weight=1)
		self.local_pattern.columnconfigure(0,weight=1)
		self.local_pattern.columnconfigure(1,weight=1)
		self.local_pattern.columnconfigure(2,weight=1)
		self.local_pattern.columnconfigure(3,weight=1)
		self.local_pattern.columnconfigure(4,weight=1)


		self.local_show_patterns = Frame(self.local_pattern)
		self.local_show_patterns.grid(column=0,row=1,sticky='nsew')
		self.local_pattern_label = Label(self.local_pattern,text="Local Patterns",font=('Times New Roman bold',12),borderwidth=5,relief=RIDGE)
		self.local_pattern_label.grid(column=0,row=0,columnspan=5,sticky='nsew')
		self.local_pattern_filter_button = Button(self.local_pattern,text='Reset Query Output',font=('Times New Roman bold',12),command=self.reset_output)
		self.local_pattern_filter_button.grid(column=0,row=2)
		self.local_pattern_filter_button = Button(self.local_pattern,text='Filter Output',font=('Times New Roman bold',12),command=self.show_updated_output)
		self.local_pattern_filter_button.grid(column=2,row=2)
		self.draw_pattern_button = Button(self.local_pattern,text='Draw Pattern',font=('Times New Roman bold',12),command=self.pop_up_pattern)
		self.draw_pattern_button.grid(column=4,row=2)

		self.local_pattern_table_frame = Frame(self.local_pattern)
		self.local_pattern_table_frame.grid(row=1,column=0,columnspan=5,sticky='nsew')
		self.local_pattern_table = Table(self.local_pattern_table_frame)
		self.local_pattern_table.show()


#---------------------------------explanation frame-----------------------------# 

		self.explanation_frame.rowconfigure(0,weight=1)
		self.explanation_frame.rowconfigure(1,weight=10)
		self.explanation_frame.rowconfigure(2,weight=1)
		self.explanation_frame.columnconfigure(0,weight=10)
		self.exp_label = Label(self.explanation_frame,text="Top Explanations",font=('Times New Roman bold',12),borderwidth=5,relief=RIDGE)
		self.exp_label.grid(column=0,row=0,sticky='nsew')
		self.exp_table_frame = Frame(self.explanation_frame)
		self.exp_table_frame.grid(row=1,column=0,sticky='nsew')
		self.exp_table = Table(self.exp_table_frame)
		self.exp_table.show()
		self.describe_exp_button = Button(self.explanation_frame,text="Describe Explanation",font=('Times New Roman bold',12),command=self.pop_up_explanation)
		self.describe_exp_button.grid(row=2,column=0)

#----------------------------------Functions----------------------------------------#

	def run_query(self):
		
		self.user_query,self.query_group_str,self.agg_function,self.user_agg,self.agg_name,self.cur_table_name=self.query_temp.get_query()
		# logger.debug(self.user_query)
		self.handle_view ="\nDROP VIEW IF EXISTS user_query;"+\
		"\nCREATE VIEW user_query as "+ self.user_query
		# logger.debug(self.handle_view)
		try:
			self.cur.execute(self.handle_view)
		except:
			tkinter.messagebox.showinfo("Info","Invalid Query, Please Doublecheck!")

		self.original_query_result_df = pd.read_sql(self.user_query,self.conn)
		self.query_result_df = self.original_query_result_df

		self.plot_data_convert_dict,self.query_data_convert_dict = self.db_info.get_db_data_type(self.cur_table_name)
		self.plot_data_convert_dict[self.agg_name] = 'numeric'
		self.query_data_convert_dict[self.agg_name] = 'float'

		self.assigned_global_table = 'dev.{}_global'.format(self.cur_table_name)
		self.assigned_local_table = 'dev.{}_local'.format(self.cur_table_name)

		model = TableModel(dataframe=self.original_query_result_df)
		self.query_result_table.updateModel(model)
		self.query_result_table.redraw()

		# if(self.cur_table_name.lower()==self.pub_dict['dict_name']):
		# 	self.table_dict = self.pub_dict
		# elif(self.cur_table_name.lower()==self.crime_dict['dict_name']):
		# 	self.table_dict = self.crime_dict


	def show_global_pattern(self):

		global_query = "select array_to_string(fixed,',') as Partition,array_to_string(variable,',') as Predictor,agg,"+\
		"round((lambda)::numeric(4,2),2) as Support,model from "+self.assigned_global_table+\
		" where array_to_string(array_sort(fixed||variable),',')='"+self.query_group_str+"';"
		# logger.debug(global_query)
		self.global_pattern_df = pd.read_sql(global_query,self.conn)
		# logger.debug(self.global_pattern_df.head())
		# logger.debug(list(self.global_pattern_df))

		pattern_model = TableModel(dataframe=self.global_pattern_df)
		self.global_pattern_table.updateModel(pattern_model)
		self.global_pattern_table.redraw()


	def show_local_pattern(self):

		local_query = "select array_to_string(fixed,',') as Partition,array_to_string(variable,',') as Predictor,"+\
		"array_to_string(fixed_value,',') as partition_values,agg,model,fixed,fixed_value,variable,"+\
		"theta,param,stats,dev_pos,dev_neg from "+self.assigned_local_table+\
		" where array_to_string(array_sort(fixed||variable),',')='"+self.query_group_str+"';"

		for n in self.local_pattern_table.multiplerowlist:
			self.chosen_local_pattern = self.local_output_pattern_df.iloc[int(n)]

		self.local_output_pattern_df = pd.read_sql(local_query,self.conn)
		self.local_output_pattern_df['stats'] = self.local_output_pattern_df['stats'].str.split(',',expand=True)[0]
		self.local_output_pattern_df['stats'] = self.local_output_pattern_df['stats'].str.strip('[')
		self.local_output_pattern_df["stats"] = pd.to_numeric(self.local_output_pattern_df["stats"])
		self.local_output_pattern_df["stats"] = self.local_output_pattern_df["stats"].round(2)

		local_shown = self.local_output_pattern_df[['partition','partition_values','predictor','agg']]

		pattern_model = TableModel(local_shown)
		self.local_pattern_table.updateModel(pattern_model)
		self.local_pattern_table.redraw()


	def use_global_filter_local(self):

		pattern_df_lists = []

		for n in self.global_pattern_table.multiplerowlist:

			model_name = self.global_pattern_df.iloc[int(n)]['model']
			# logger.debug("model_name"+model_name)
			global_partition = self.global_pattern_df.iloc[int(n)]['partition']
			global_predictor = self.global_pattern_df.iloc[int(n)]['predictor']

			g_filter_l_query = " select array_to_string(fixed,',') as Partition,array_to_string(variable,',') as Predictor,"+\
			"array_to_string(fixed_value,',') as partition_values,agg,model,fixed,fixed_value,variable,"+\
			"theta,param,stats,dev_pos,dev_neg from "+self.assigned_local_table+\
			" where array_to_string(fixed,',')='"+global_partition+\
			"' and array_to_string(variable,',')='"+global_predictor+\
			"' and model = '"+model_name+"';"

			# logger.debug(g_filter_l_query)

			self.local_output_pattern_df = pd.read_sql(g_filter_l_query,self.conn)
			self.local_output_pattern_df['stats'] = self.local_output_pattern_df['stats'].str.split(',',expand=True)[0]
			self.local_output_pattern_df['stats'] = self.local_output_pattern_df['stats'].str.strip('[')
			self.local_output_pattern_df["stats"] = pd.to_numeric(self.local_output_pattern_df["stats"])
			self.local_output_pattern_df["stats"] = self.local_output_pattern_df["stats"].round(2)

			local_shown = self.local_output_pattern_df[['partition','partition_values','predictor','agg']]

		model = TableModel(dataframe=local_shown)
		self.local_pattern_table.updateModel(model)
		self.local_pattern_table.redraw()


	def global_description(self):

		for n in self.global_pattern_table.multiplerowlist:
			fixed_attribute = self.global_pattern_df.iloc[int(n)]['partition']
			aggregation_function=self.global_pattern_df.iloc[int(n)]['agg']
			modeltype = self.global_pattern_df.iloc[int(n)]['model']
			variable_attribute = self.global_pattern_df.iloc[int(n)]['predictor']
			Lambda = self.global_pattern_df.iloc[int(n)]['support']

		fixed_attribute=fixed_attribute.replace(",",", ")

		global_desc = "For each ("+fixed_attribute+'), the '+aggregation_function +' is '+modeltype+'\n in '+variable_attribute+'.'+\
		'This pattern holds for '+str(Lambda*100)+ ' % of the '+fixed_attribute
		desc_win = Toplevel()
		x = self.parent.winfo_x()
		y = self.parent.winfo_y()
		w = 540
		h = 120 
		desc_win.geometry("%dx%d+%d+%d" % (w, h, x + 450, y + 500))
		desc_win.wm_title("Global Pattern Description")
		desc_frame = Frame(desc_win)
		desc_frame.pack(fill=BOTH,expand=True)
		desc_label= Label(desc_frame,text=global_desc,font=('Times New Roman bold',12),borderwidth=5,relief=SOLID,justify=LEFT)
		desc_label.pack(fill=BOTH,expand=True)


	def use_local_filter_output(self):  # given partition attributes and partition values, get explanation(query on user query)
		l_filter_o_query = None

		for n in self.local_pattern_table.multiplerowlist:

			chosen_row = self.local_output_pattern_df.iloc[int(n)]
			partition_attr_list = self.local_output_pattern_df.iloc[int(n)]['partition'].split(',')
			partition_value_list = self.local_output_pattern_df.iloc[int(n)]['partition_values'].split(',')

		where_clause_list = []
		where_clause = None

		for n in range(len(partition_attr_list)):
			if(self.query_data_convert_dict[partition_attr_list[n]]=='str'):
					condition = "{} =\'{}\'".format(partition_attr_list[n],partition_value_list[n])
			elif(self.query_data_convert_dict[partition_attr_list[n]]=='float'):
				condition = "{} = {}::float".format(partition_attr_list[n],partition_value_list[n])
			else:
				condition = "{} = {}::int".format(partition_attr_list[n],partition_value_list[n])

			where_clause_list.append(condition)

		if(len(where_clause_list)==1):
			where_clause = where_clause_list[0]
		else:
			where_clause = " and ".join(where_clause_list)
				
		l_filter_o_query = "select user_query.* from user_query where "+where_clause+";"
		# logger.debug("filter_output_query:")
		# logger.debug(l_filter_o_query)
		filtered_result_df = pd.read_sql(l_filter_o_query,self.conn)

		return chosen_row,filtered_result_df


	def show_updated_output(self):

		filtered_result_df = self.use_local_filter_output()[1]
		self.query_result_df = filtered_result_df
		model = TableModel(dataframe=filtered_result_df)
		self.query_result_table.updateModel(model)
		self.query_result_table.redraw()


	def handle_question(self,direction):

		self.question_tuple = ''
		config=ExplConfig()
		config.conn = self.config.conn
		config.cur = self.config.cur
		config.query_table_name = self.cur_table_name
		eg = ExplanationGenerator(config, {'pattern_table':'dev.{}'.format(self.cur_table_name),'query_result_table':self.cur_table_name})
		eg.initialize() 
		col_name = ['Explanation_Tuple',"Score",'From_Pattern',"Drill_Down_To","Distance","Outlierness","Denominator","relevent_model","relevent_param","refinement_model","drill_param"]
		exp_df = pd.DataFrame(columns=["From_Pattern","Drill_Down_To","Score","Distance","Outlierness","Denominator","relevent_model","relevent_param","refinement_model","drill_param"])
		for n in self.query_result_table.multiplerowlist:
			self.question = self.query_result_df.iloc[[int(n)]]
			self.original_question = self.question.copy(deep=True)

			self.question.rename(columns={self.agg_name:self.user_agg}, inplace=True)
			self.question_tuple = self.query_result_df.iloc[[int(n)]]
			# logger.debug(self.question)
			self.question['direction']=direction
			self.question['lambda'] = 0.2
			question = self.question.iloc[0].to_dict()
			# logger.debug(question)
			elist = eg.do_explain_online(question)

			exp_list=[]
			for e in elist:
				tuple_list=[]
				# print(str(e.tuple_value))
				# print(str(e.tuple_value.keys()))
				# e_tuple_str = ','.join(map(str, e.tuple_value.values()))
				e_tuple_str = e.ordered_tuple_string()
				tuple_list.append(e_tuple_str)

				score = round(e.score,2)
				tuple_list.append(score)

				if e.expl_type == 1:
					local_pattern=(
						'[' + ','.join(e.relevent_pattern[0]) +\
						'=' + ','.join(list(map(str, e.relevent_pattern[1]))) +']:'+ \
						','.join(list(map(str, e.relevent_pattern[2])))+' \u2933 '+self.agg_name
						)
					relevent_model = e.relevent_pattern[4]
					if e.relevent_pattern[4] == 'const':						
						relevent_param = str(round(float(e.relevent_pattern[6].split(',')[0][1:]),2))
					else:
						# relevent_param = 'Intercept=' + str(e.relevent_pattern[7]['Intercept'])+', '+str(list(e.relevent_pattern[7])[1])+'='+str(round(e.relevent_pattern[7][list(e.relevent_pattern[7])[1]],2))
						relevent_param = e.relevent_pattern[7]

					drill_down_to = ','.join([x for x in e.refinement_pattern[0] if x not in e.relevent_pattern[0]])
					refinement_model = e.refinement_pattern[4]
					if e.refinement_pattern[4] == 'const':

						drill_param =str(round(float(e.refinement_pattern[6].split(',')[0][1:]),2))
					else:
						drill_param =e.refinement_pattern[7]
				else:
					relevent_model = e.relevent_pattern[4]
					local_pattern=(
						'[' + ','.join(e.relevent_pattern[0]) +\
						'=' + ','.join(list(map(str, e.relevent_pattern[1]))) +']:'+ \
						','.join(list(map(str, e.relevent_pattern[2])))+' \u2933 '+self.agg_name
						)
					if e.relevent_pattern[4] == 'const':
						relevent_param = str(round(float(e.relevent_pattern[6].split(',')[0][1:]),2))
					else:
						# relevent_param = 'Intercept=' + str(e.relevent_pattern[7]['Intercept'])+', '+str(list(e.relevent_pattern[7])[1])+'='+str(e.relevent_pattern[7][list(e.relevent_pattern[7])[1]])
						relevent_param = e.relevent_pattern[7]

					refinement_model = ''
					drill_down_to = ''
					drill_param = ''
				tuple_list.append(local_pattern)
				tuple_list.append(drill_down_to)
				distance = round(e.distance,2)
				tuple_list.append(distance)
				outlierness = round(e.deviation,2)
				tuple_list.append(outlierness)
				denominator = round(e.denominator,2)
				tuple_list.append(denominator)
				tuple_list.append(relevent_model)
				tuple_list.append(relevent_param)
				tuple_list.append(refinement_model)
				tuple_list.append(drill_param)
				exp_list.append(tuple_list)

			df_exp = pd.DataFrame(exp_list,columns=col_name)
			exp_df = exp_df.append(df_exp,ignore_index=True)
		
		self.exp_df = exp_df[col_name]
		model = TableModel(dataframe=self.exp_df)
		self.exp_table.updateModel(model)	
		self.exp_table.redraw()

	def handle_low(self):
		self.user_direction='low'
		self.handle_question(self.user_direction)

	def handle_high(self):
		self.user_direction='high'
		self.handle_question(self.user_direction)


	def reset_output(self):

		model = TableModel(dataframe=self.original_query_result_df)
		self.query_result_table.updateModel(model)	
		self.query_result_table.redraw()

		self.query_result_df = self.original_query_result_df


	def pop_up_pattern(self):

		chosen_row,pattern_data_df = self.use_local_filter_output()
		self.local_pattern_detail = Local_Pattern_Frame(chosen_row=chosen_row,pattern_data_df=pattern_data_df,
			agg_alias=self.agg_name,data_convert_dict=self.plot_data_convert_dict)
		
		self.local_pattern_detail.load_pattern_description()
		self.local_pattern_detail.load_pattern_graph()


	def get_pattern_result(self,partition_attr_list=None,partition_value_list=None,pred_attr_list=None):  # given partition attributes and partition values, get explanation(query on table)

		where_clause_list = []
		where_clause = None

		# logger.debug("partition_attr_list is ")
		# logger.debug(partition_attr_list)

		# logger.debug("partition_value_list is ")
		# logger.debug(partition_value_list)


		for n in range(len(partition_attr_list)):
			if(self.query_data_convert_dict[partition_attr_list[n]]=='str'):
					condition = "{} =\'{}\'".format(partition_attr_list[n],partition_value_list[n])
			elif(self.query_data_convert_dict[partition_attr_list[n]]=='float'):
				condition = "{} = {}::float".format(partition_attr_list[n],partition_value_list[n])
			else:
				condition = "{} = {}::int".format(partition_attr_list[n],partition_value_list[n])
			where_clause_list.append(condition)

		if(len(where_clause_list)==1):
			where_clause = where_clause_list[0]
		else:
			where_clause = " and ".join(where_clause_list)

		if (pred_attr_list is not None):

			Pattern_Q = "SELECT "+self.agg_function+" as "+self.agg_name+","+','.join(partition_attr_list)+","+','.join(pred_attr_list)+\
			" FROM " +self.cur_table_name+" WHERE " + where_clause+\
			" GROUP BY "+','.join(partition_attr_list)+","+','.join(pred_attr_list)
		else:
			Pattern_Q = "SELECT "+self.agg_function+" as "+self.agg_name+","+','.join(partition_attr_list)+\
			" FROM " +self.cur_table_name + " WHERE " + where_clause+\
			" GROUP BY "+','.join(partition_attr_list)


		# logger.debug("Pattern_Q")
		# logger.debug(Pattern_Q)

		exp_pattern_df = pd.read_sql(Pattern_Q,self.conn)

		# logger.debug('exp_pattern_df is :')
		# logger.debug(exp_pattern_df)

		return exp_pattern_df



	def pop_up_explanation(self):

		for n in self.exp_table.multiplerowlist:
			exp_chosen_row = self.exp_df.iloc[int(n)]
			relevent_pattern = self.exp_df.iloc[int(n)]['From_Pattern']
			rel_pattern_part = relevent_pattern.split(':')[0].split('=')[0].strip('[')
			rel_pattern_part_value = relevent_pattern.split(':')[0].split('=')[1].split(']')
			rel_pattern_pred = relevent_pattern.split(':')[1].split(' \u2933 ')[0]
			agg_name = relevent_pattern.split(':')[1].split(' \u2933 ')[1]
			rel_pattern_model = self.exp_df.iloc[int(n)]['relevent_model']
			rel_param = self.exp_df.iloc[int(n)]['relevent_param']
			rel_pattern_part_list = rel_pattern_part.split(',')
			rel_pattern_pred_list = rel_pattern_pred.split(',')
			rel_pattern_part_value_list = rel_pattern_part_value[0].split(',')
			exp_tuple = self.exp_df.iloc[int(n)]['Explanation_Tuple']
			exp_tuple_list = exp_tuple.split(',')[:-1]
			exp_tuple_score = float(self.exp_df.iloc[int(n)]['Score'])
			drill_attr_list = self.exp_df.iloc[int(n)]['Drill_Down_To'].split(',')

			if(drill_attr_list!=['']):
				exp_tuple_col = rel_pattern_part_list + rel_pattern_pred_list + drill_attr_list
				exp_tuple_col.sort()
			else:
				exp_tuple_col = rel_pattern_part_list + rel_pattern_pred_list
				exp_tuple_col.sort()


		# logger.debug('exp_tuple_col is:')
		# logger.debug(exp_tuple_col)

		# logger.debug('exp_tuple_list is:')
		# logger.debug(exp_tuple_list)


		exp_tuple_df_list = [exp_tuple_list]
		exp_tuple_df = pd.DataFrame(exp_tuple_df_list)
		logger.debug("exp_tuple_df:")
		exp_tuple_df.columns = exp_tuple_col
		logger.debug(exp_tuple_df)


		if(drill_attr_list != ['']):

			drill_values = []
			for n in drill_attr_list:
				drill_value = exp_tuple_df[n].to_string(index=False)
				drill_values.append(drill_value)

			# logger.debug('rel_pattern_part_list')
			# logger.debug(rel_pattern_part_list)

			# logger.debug('rel_pattern_part_value_list')
			# logger.debug(rel_pattern_part_value_list)


			drill_pattern_df = self.get_pattern_result(partition_attr_list=rel_pattern_part_list+drill_attr_list,
													 partition_value_list=rel_pattern_part_value_list+drill_values,pred_attr_list=rel_pattern_pred_list)

			# logger.debug("drill_pattern_df is")
			# logger.debug(drill_pattern_df)

			rel_pattern_df = self.get_pattern_result(partition_attr_list=rel_pattern_part_list,
													 partition_value_list=rel_pattern_part_value_list,pred_attr_list=rel_pattern_pred_list)

			question_df = self.original_question

			explanation_df = self.get_pattern_result(partition_attr_list=exp_tuple_col,
													 partition_value_list=exp_tuple_list,pred_attr_list=None)
			logger.debug(explanation_df)
			exp_selected = exp_chosen_row

			data_convert_dict = self.plot_data_convert_dict

			self.Explainer = Exp_Frame(input_question_df=question_df, input_explanation_df=explanation_df, input_exp_chosen_row=exp_selected, input_none_drill_down_df=rel_pattern_df,
				input_drill_down_df=drill_pattern_df, input_data_convert_dict=data_convert_dict)

			self.Explainer.load_exp_graph()
			self.Explainer.load_exp_description(user_direction=self.user_direction)

		else:
			rel_pattern_df = self.get_pattern_result(partition_attr_list=rel_pattern_part_list,
														 partition_value_list=rel_pattern_part_value_list,pred_attr_list=rel_pattern_pred_list)

			question_df = self.original_question

			explanation_df = self.get_pattern_result(partition_attr_list=exp_tuple_col,
													 partition_value_list=exp_tuple_list,pred_attr_list=None)
			exp_selected = exp_chosen_row

			data_convert_dict = self.plot_data_convert_dict

			self.Explainer = Exp_Frame(input_question_df=question_df, input_explanation_df=explanation_df, input_exp_chosen_row=exp_selected, input_none_drill_down_df=rel_pattern_df,
				input_drill_down_df=None, input_data_convert_dict=data_convert_dict)

			self.Explainer.load_exp_graph()
			self.Explainer.load_exp_description(user_direction=self.user_direction)

def startCapeGUI(conn,config):
	root = Tk()
	root.title('CAPE')
	default_font = nametofont("TkTextFont")
	default_font.configure(size=12)
	bigfont = Font(family="Helvetica",size=12)
	root.option_add('*TCombobox*Listbox.font',bigfont)
	root.option_add('*TCombobox*Listbox.selectBackground', 'wheat1')
	width, height = root.winfo_screenwidth(), root.winfo_screenheight()
	root.geometry('%dx%d+0+0' % (width,height))
	ui = CAPE_UI(root,conn=conn,config=config)
	root.mainloop()


def main():

	config=DictLike()
	dbconn=DBConnection(host="127.0.0.1",user="antiprov",db="antiprov",password='1234',port="5432")
	conn=dbconn.pgconnect()
	startCapeGUI(conn=conn,config=config)

if __name__ == '__main__':
	main()







"""
Data structures for representing aggregate regression patterns (ARPs).
"""

"""
Represents a global pattern (F, V, g)
"""
class GlobalPattern(object):
    
    def __init__(self, f, v, model, theta, lamb):
        self.f = f
        self.v = v
        self.model = model
        self.theta = theta
        self.lamb = lamb
        

class LocalPattern(object):
    
    def __init__(self, f, f_value, v, model, theta):
        self.f = f
        self.f_value = f_value
        self.v = v
        self.model = model
        self.theta = theta

class SetTrie(object):

    def __init__(self):
        self.root = TrieEntry()
        
    def search(self, attr):
        """
        :param attr: list of sorted int
        :return: 2 lists of patterns
        """
        cur=self.root
        for i in attr:
            if i not in cur.child:
                return None, None
            cur=cur.child[i]
        else:
            return cur.l, cur.g

    def search_subset(self, attr):
        loc = []
        glob = []
        self.__search_subset(self.root, attr, 0, loc, glob)
        return loc, glob

    def __search_subset(self, te, attr, start, loc, glob):
        for i in range(start,len(attr)):
            if attr[i] in te.child:
                ne = te.child[attr[i]]
                if ne.l:
                    loc+=ne.l
                if ne.g:
                    glob+=ne.g
                self.__search_subset(ne, attr, i+1, loc, glob)


class TrieEntry(object):
    """
        Attributes:
            child: dictionary that points to the group that has G
                      as prefix
            g: list of global patterns under the group
            l: list of local patterns under the group
    """
    def __init__(self):
        self.g = None
        self.l = None
        self.child = {}


class PatternCollection(object):
    """
        Attributes:
            header: list of attribute names
            index: dict that maps attribute names to their index in header
            a: list of dictionaries. Each index is an attr, and each
               dictionary maps an aggregate function to the root of the
               trie that stores patterns
    """
    
    def __init__(self, header):
        """
        type header: list of strings
        """
        self.header = header
        self.index={}
        for i in range(len(header)):
            self.index[header[i]] = i
        self.a = [{} for i in range(len(header))]
        
    def search(self, a, attr, agg):
        """
        type a: string
        type attr: list of strings
        type agg: string
        """
        return self.a[self.index[a]][agg].search([self.index[i] for i in attr].sort())

    def search_subset(self, a, attr, agg):

        return self.a[self.index[a]][agg].search_subset([self.index[i] for i in attr].sort())

    def add_global(self, f, v, a, agg, model, theta, lamb):
        """      
            type f: list of strings
            type v: list of strings
            type a: string
            type agg: string
            model: string
            theta: float
            lamb: float
        """
        a=self.index[a]
        f=[self.index[i] for i in f]
        v=[self.index[j] for j in v]
        attr=sorted(f+v)
        if agg not in self.a[a]:
            self.a[a][agg]=SetTrie()
        cur=self.a[a][agg].root
        for i in attr:
            if i not in cur.child:
                cur.child[i]=TrieEntry()
            cur=cur.child[i]
        if not cur.g:
            cur.g=[]
        cur.g.append(GlobalPattern(f,v,model,theta,lamb))
    
    def add_local(self, f, f_value, v, a, agg, model, theta):
        a=self.index[a]
        f=[self.index[i] for i in f]
        v=[self.index[j] for j in v]
        attr=sorted(f+v)
        if agg not in self.a[a]:
            self.a[a][agg]=SetTrie()
        cur=self.a[a][agg].root
        for i in attr:
            if i not in cur.child:
                cur.child[i]=TrieEntry()
            cur=cur.child[i]
        if not cur.l:
            cur.l=[]
        cur.l.append(LocalPattern(f,f_value,v,model,theta))
    



#!/usr/bin/python
# -*- coding:utf-8 -*- 

import sys, getopt
import pandas
import csv
import time
import math

from capexplain.utils import *

def compare(t1, t2, aggr_col):
    for k in range(0, len(t1)):
        if t1[k] != t2[k]:
            if t1[k] < t2[k]:
                return -1
            else:
                return 1
    return 0

class CategoryEmbedding(object):
    """Summary of class here.
    """

    def __compute_distance(self, col, val1, val2, aggr_col):
        if col not in self.cate_cols:
            return -1
        else:
            if val1 not in self.cate_cols[col] or val2 not in self.cate_cols[col]:
                return -1
            len1 = len(self.cate_cols[col][val1])
            len2 = len(self.cate_cols[col][val2])
            i = 0
            j = 0
            dist = 0.0
            cnt = 0
            while i < len1 and j < len2:
                cmp_res = compare(self.cate_cols[col][val1][i][0], self.cate_cols[col][val2][j][0], aggr_col)
                cnt += 1
                if cmp_res == 0:
                    dist += (self.cate_cols[col][val1][i][1] - self.cate_cols[col][val2][j][1]) * (self.cate_cols[col][val1][i][1] - self.cate_cols[col][val2][j][1])
                    i += 1
                    j += 1
                elif cmp_res < 0:
                    dist += self.cate_cols[col][val1][i][1] * self.cate_cols[col][val1][i][1]
                    i += 1
                else:
                    dist += self.cate_cols[col][val2][j][1] * self.cate_cols[col][val2][j][1]
                    j += 1
            while i < len1:
                dist += self.cate_cols[col][val1][i][1] * self.cate_cols[col][val1][i][1]
                i += 1
                cnt += 1
            while j < len2:
                dist += self.cate_cols[col][val2][j][1] * self.cate_cols[col][val2][j][1]
                j += 1
                cnt += 1
            return math.sqrt(dist / cnt)

    def __init__(self, df=[], aggr_col='count'):
        """Inits SampleClass with blah."""
        self.df = df
        self.cate_val_mark = {}
        self.cate_val_cnt = {}
        self.cate_val_list = {}
        self.cate_cols = {}
        self.dist_map = {}
        self.cos_sim_map = {}
        self.dist_stats = {}
        col_list = []
        for col in df:
            if col == 'index' or col == aggr_col:
                continue
            if df[col].dtype.kind == 'S' or df[col].dtype.kind == 'O':
                col_list.append(col)
                self.cate_val_mark[col] = {}
                self.cate_val_cnt[col] = 0
                self.cate_val_list[col] = []
                

        for idx, row in df.iterrows():
            for col in col_list:
                if row[col] not in self.cate_val_mark[col]:
                    self.cate_val_mark[col][row[col]] = self.cate_val_cnt[col]
                    self.cate_val_list[col].append(row[col])
                    self.cate_val_cnt[col] += 1

        for col in col_list:
            self.dist_map[col] = [[0.0 for i in range(self.cate_val_cnt[col])] for j in range(self.cate_val_cnt[col])]
            # self.cos_sim_map[col] = {}
            self.dist_stats[col] = {'max':0.0, 'min':1e10}

        print col_list
        print self.cate_val_list['pubkey']
        for col in df:
            if col == 'index' or col == 'name':
                continue
            if df[col].dtype.kind == 'S' or df[col].dtype.kind == 'O':
            #if is_string_dtype(df[col]):
                temp_col_list = col_list
                temp_col_list.remove(col)
                #print temp_col_list, col
                res = df.sort([col] + temp_col_list)
                self.cate_cols[col] = {}
                for idx, row in res.iterrows():
                    t = [map(lambda x: self.cate_val_mark[x][row[x]], temp_col_list), row[aggr_col]]
                    # for col2 in temp_col_list:
                    #     t[0].append(self.cate_val_mark[col2][row[col2]])
                    if not row[col] in self.cate_cols[col]:
                        #self.cate_cols[col][row[col]] = [(projection(row, temp_col_list), row[aggr_col])]
                        self.cate_cols[col][row[col]] = [t]
                    else:
                        #self.cate_cols[col][row[col]].append((projection(row, temp_col_list), row[aggr_col]))
                        if compare(t, self.cate_cols[col][row[col]][-1][0], aggr_col) != 0:
                            self.cate_cols[col][row[col]].append([t, row[aggr_col]])
                        else:
                            self.cate_cols[col][row[col]][-1][1] += row[aggr_col]
                temp_col_list.append(col)

        print self.cate_val_cnt
        for col in self.cate_cols:
            print col
            cnt = 0
            for val1 in self.cate_cols[col]:
                for val2 in self.cate_cols[col]:
                    id1 = self.cate_val_mark[col][val1]
                    id2 = self.cate_val_mark[col][val2]
                    cnt += 1
                    if val1 != val2:
                        self.dist_map[col][id1][id2] = self.__compute_distance(col, val1, val2, aggr_col)
                        if self.dist_map[col][id1][id2] > self.dist_stats[col]['max']:
                            self.dist_stats[col]['max'] = self.dist_map[col][id1][id2]
                        if self.dist_map[col][id1][id2] < self.dist_stats[col]['min']:
                            self.dist_stats[col]['min'] = self.dist_map[col][id1][id2]
                        # self.cos_sim_map[col][val1][val2] = self.__compute_cos_similarity(col, val1, val2, aggr_col)
                    else:
                        self.dist_map[col][id1][id2] = 0
                        # self.cos_sim_map[col][val1][val2] = 1.0

                    if cnt > 1000000:
                        cnt = 0
                        print col, val1, val2

            for val1 in self.cate_cols[col]:
                for val2 in self.cate_cols[col]:
                    if val1 != val2:
                        id1 = self.cate_val_mark[col][val1]
                        id2 = self.cate_val_mark[col][val2]
                        self.dist_map[col][id1][id2] -= 0.9 * self.dist_stats[col]['min']

    def compute_similarity(self, col, val1, val2, aggr_col):
        if col not in self.dist_map:
            return -1
        else:
            if val1 not in self.cate_val_mark[col] or val2 not in self.cate_val_mark[col]:
                return -1
            id1 = self.cate_val_mark[col][val1]
            id2 = self.cate_val_mark[col][val2]
            if val1 == val2:
                return 1.0
            else:
                
                return (1 - (self.dist_map[col][id1][id2]) 
                        / (self.dist_stats[col]['max']))
  
    def is_categorical(self, col):
        return col in self.cate_cols



    



"""
get information of databases for corresponding connection
"""
import psycopg2
import pandas as pd 
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)

formatter = logging.Formatter('%(asctime)s %(levelname)s line %(lineno)d: %(message)s')

stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)

logger.addHandler(stream_handler)

# logging.basicConfig(level=logging.DEBUG,
# 	format='%(asctime)s %(levelname)s line %(lineno)d: %(message)s')


class DBinfo:

	def __init__(self,conn=None):
		self.conn = conn

	def get_table_name(self):  # get table names in current connection

		extract_table_query = """
							  SELECT table_name
							  FROM information_schema.tables
							  WHERE table_schema='public'
							  AND table_type='BASE TABLE';
		                    """
		df_tables = pd.read_sql(extract_table_query, self.conn)

		return (df_tables['table_name'].values.tolist())

	def get_attribute(self,table_name):  # return the attributes of a table as a list

		extract_table_attrs_query = "select CONCAT_WS (' : ', column_name , data_type) as attribute from information_schema.columns where table_name =" + '\''+table_name +'\''
		df_attributes = pd.read_sql(extract_table_attrs_query,self.conn)

		return (df_attributes['attribute'].values.tolist())

	def get_db_info(self):

		extract_table_query = """
							  SELECT table_name
							  FROM information_schema.tables
							  WHERE table_schema='public'
							  AND table_type='BASE TABLE';
		                    """
		df_tables = pd.read_sql(extract_table_query, self.conn)
		df_list = df_tables['table_name'].values.tolist()
		table_dict = dict.fromkeys(df_list,0)
		for table in df_list:
			extract_table_attrs_query = "select column_name as attribute from information_schema.columns where table_name = " + '\''+table+'\''
			table_name_attr = pd.read_sql(extract_table_attrs_query,self.conn)
			table_name_attr_list = table_name_attr['attribute'].values.tolist()
			table_dict[table] = table_name_attr_list

		return table_dict

	def get_db_data_type(self,table_name):

		pg_numeric_list = ['smallint','integer','bigint','decimal','numeric','real','double precision','smallserial','serial','bigserial']
		pg_int_list = ['smallint','integer','bigint','smallserial','serial','bigserial']
		pg_float_list = ['decimal','numeric','real','double precision']
		data_type_query = "SELECT column_name, data_type FROM information_schema.columns"+\
		                  " WHERE table_name = \'"+table_name+"\';"
		data_type_df = pd.read_sql(data_type_query,self.conn)
		logger.debug(data_type_df)
		data_type_list = data_type_df.values.tolist()
		column_name_list = data_type_df.column_name.values.tolist()

		query_data_convert_dict = dict.fromkeys(column_name_list,None)
		plot_data_convert_dict = dict.fromkeys(column_name_list, None)


		for n in data_type_list:
			if(n[1] in pg_numeric_list):
				plot_data_convert_dict[n[0]] = 'numeric'
				if(n[1] in pg_int_list):
					query_data_convert_dict[n[0]]='int'
				else:
					query_data_convert_dict[n[0]]='float'
			else:
				query_data_convert_dict[n[0]] = 'str'
				plot_data_convert_dict[n[0]]='str'

		logger.debug(plot_data_convert_dict)
		logger.debug(query_data_convert_dict)


		return plot_data_convert_dict,query_data_convert_dict


if __name__ == "__main__":
	conn = psycopg2.connect(dbname="antiprov",user="antiprov",host="127.0.0.1",port="5432",password='1234')
	dbinfo = DBinfo(conn)

	dbinfo.get_db_data_type('pub')
















"""
define the plotting functions used in "CAPE"
"""
import matplotlib
from matplotlib.figure import Figure
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.mplot3d import Axes3D
import pandas as pd
from math import floor,ceil
import re
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)s %(levelname)s %(name)s line %(lineno)d: %(message)s')
stream_handler = logging.StreamHandler()
stream_handler.setFormatter(formatter)
logger.addHandler(stream_handler)



coded = re.compile('(coded\_).*')

class Plotter:

	def __init__(self,figure,data_convert_dict,mode='2D'):
		
		self.data_convert_dict = data_convert_dict
		self.figure = figure
		self.legend_proxies_shape= []  # for 3d legend specifically
		self.legend_proxies_name = [] # for 3d legend specifically

		if(mode=='2D'):
			self.a = self.figure.add_subplot(111)
		elif(mode=='3D'):
			self.a= self.figure.gca(projection='3d')

		self.x_max = None
		self.x_min = None
		self.y_max = None
		self.y_min = None
		self.z_max = None
		self.z_min = None

		self.encoding_list = []


	def df_type_conversion(self,data_frame):  # convert column data types based on database information(plot_data_convert_dict)
		numeric_cols = []
		str_cols = []

		for n in list(data_frame):
			if coded.search(n) is not None:
				continue
			elif(self.data_convert_dict[n]=='numeric'):
				numeric_cols.append(n) 
			elif(self.data_convert_dict[n]=='str'):
				str_cols.append(n)

		if(len(numeric_cols)>0):
			data_frame[numeric_cols] = data_frame[numeric_cols].apply(lambda x : pd.to_numeric(x))

		if(len(str_cols)>0):
			for n in str_cols:
				if('coded_'+n in data_frame.columns):
					continue

				if(len(self.encoding_list)>0):
					for x in self.encoding_list:
						if(('coded_'+n) == x['dict_name']):
							data_frame['coded_'+n] = data_frame[n].map(x)

				if(('coded_'+n) not in data_frame.columns):
					data_frame['coded_'+n] = data_frame[n].astype('category').cat.codes
					dict1 = dict(zip(data_frame[n],data_frame['coded_'+n]))
					dict1['dict_name'] = 'coded_'+n
					self.encoding_list.append(dict1)

		return data_frame
		
	def set_x_label(self,label_name):
		self.a.set_xlabel(label_name)


	def set_y_label(self,label_name):
		self.a.set_ylabel(label_name)


	def set_z_label(self,label_name):
		self.a.set_zlabel(label_name)


	def set_title(self,title_name):
		self.a.set_title(title_name)


	def plot_2D_const(self,const_value,label=None):

		self.a.axhline(const_value,c="red",linewidth=2,label=label)
		self.a.legend(loc='best')


	def plot_3D_const(self,df,x=None,y=None,z_value=None,label=None): # df is the dataframe containing the columns to be drawn

		df = self.df_type_conversion(df)

		if ('coded_'+x) in df.columns:
			x_range=np.arange(df['coded_'+x].min(),df['coded_'+x].max()+1,1)
		else:
			x_range=np.arange(df[x].min(),df[x].max()+1,1)

		if ('coded_'+y) in df.columns:
			y_range=np.arange(df['coded_'+y].min(),df['coded_'+y].max()+1,1)
		else:
			y_range=np.arange(df[y].min(),df[y].max()+1,1)

		X1, Y1 = np.meshgrid(x_range, y_range)
		zs = np.array([z_value for x1,y1 in zip(np.ravel(X1), np.ravel(Y1))])
		Z = zs.reshape(X1.shape)
		self.a.plot_surface(X1, Y1, Z,color='r',linewidth=2,label=label)

		if ('coded_'+x) in df.columns:
			self.a.set(xticks=df['coded_'+x].values, xticklabels=df[x])

		else:
			x_min = floor(df[x].values.min())
			x_max = ceil(df[x].values.max()+1)

			if(self.x_min==None):
				self.x_min=x_min
			else:
				self.x_min=min(self.x_min,x_min)

			if(self.x_max==None):
				self.x_max=x_max
			else:
				self.x_max=max(self.x_max,x_max)

			x_division_value = (self.x_max - self.x_min) // 10
			if(x_division_value==0):
				x_division_value = 1
			self.a.set(xticks=list(range(self.x_min,self.x_max,x_division_value)))

		if ('coded_'+y) in df.columns:
			self.a.set(yticks=df['coded_'+y].values, yticklabels=df[y])

		else:
			y_min = floor(df[y].values.min())
			y_max = ceil(df[y].values.max()+1)

			if(self.y_min==None):
				self.y_min=y_min
			else:
				self.y_min=min(self.y_min,y_min)

			if(self.y_max==None):
				self.y_max=y_max
			else:
				self.y_max=max(self.y_max,y_max)

			y_division_value = (self.y_max - self.y_min) // 10
			if(y_division_value==0):
				y_division_value = 1
			self.a.set(yticks=list(range(self.y_min,self.y_max,y_division_value)))


		plane_proxy_shape = plt.Rectangle((0, 0), 1, 1, fc="r")
		plane_proxy_name = label
		self.legend_proxies_shape.append(plane_proxy_shape)
		self.legend_proxies_name.append(plane_proxy_name)
		self.a.legend(self.legend_proxies_shape,self.legend_proxies_name)


	def plot_2D_linear(self,x,slope,intercept,label=None):  # x:df column; slope,intercept:numeric values
		var_min = pd.to_numeric(x.min()).item()
		var_max = pd.to_numeric(x.max()).item()
		logger.debug("var_min is:")
		logger.debug(var_min)

		logger.debug("var_max is:")
		logger.debug(var_max)

		X1 = np.linspace(var_min-1,var_max+1,100)
		X = x.astype('float64')
		y_vals = slope * X1 + intercept

		logger.debug("y_vals is:")
		logger.debug(y_vals)

		self.a.plot(X1, y_vals, c='r',linewidth=2,label=label)
		self.a.legend(loc='best')

	def plot_3D_linear(self,x,y,x_weight,y_weight):
		pass 


	def plot_2D_scatter(self,df,x=None,y=None,color='g',marker='o',size=60,zorder=0,alpha=1,label=None): # x,y are 2 df column names

		df = self.df_type_conversion(df)

		logger.debug(df)
		if ('coded_'+x) in df.columns:
			X=df['coded_'+x]
		else:
			X=df[x]

		if ('coded_'+y) in df.columns:
			Y=df['coded_'+y]
		else:
			Y=df[y]

		self.a.scatter(X, Y, s=size, c=color,marker=marker,zorder=zorder,alpha=1,label=label)
		self.a.legend(loc='best')


		if ('coded_'+x) in df.columns:

			x_ticks = df['coded_'+x].values
			self.a.set(xticks=x_ticks, xticklabels=df[x])

		else:
			x_min = floor(df[x].values.min())
			x_max = ceil(df[x].values.max()+1)

			if(self.x_min==None):
				self.x_min=x_min
			else:
				self.x_min=min(self.x_min,x_min)

			if(self.x_max==None):
				self.x_max=x_max
			else:
				self.x_max=max(self.x_max,x_max)

			x_division_value = (self.x_max - self.x_min) // 10
			if(x_division_value==0):
				x_division_value = 1
			self.a.set(xticks=list(range(self.x_min,self.x_max,x_division_value)))

		if ('coded_'+y) in df.columns:			
			y_ticks = df['coded_'+y].values
			self.a.set(yticks=y_ticks, yticklabels=df[y])

		else:
			y_min = floor(df[y].values.min())
			y_max = ceil(df[y].values.max()+1)

			if(self.y_min==None):
				self.y_min=y_min
			else:
				self.y_min=min(self.y_min,y_min)

			if(self.y_max==None):
				self.y_max=y_max
			else:
				self.y_max=max(self.y_max,y_max)

			y_division_value = (self.y_max - self.y_min) // 10
			if(y_division_value==0):
				y_division_value = 1
			self.a.set(yticks=list(range(self.y_min,self.y_max,y_division_value)))


	def plot_3D_scatter(self,df,x,y,z,color='g',marker='o',size=60,zorder=0,alpha=1,label=None): # x,y,z are 3 df columns

		df = self.df_type_conversion(df)

		if ('coded_'+x) in df.columns:
			X=df['coded_'+x]
		else:
			X=df[x]

		if ('coded_'+y) in df.columns:
			Y=df['coded_'+y]
		else:
			Y=df[y]

		if ('coded_'+z) in df.columns:
			Z=df['coded_'+z]
		else:
			Z=df[z]

		self.a.scatter(X, Y, Z, s=size, c=color,marker=marker,zorder=zorder,alpha=alpha,label=label)
		
		if ('coded_'+x) in df.columns:
			x_ticks = df['coded_'+x].values
			self.a.set(xticks=x_ticks, xticklabels=df[x])


		else:
			x_min = floor(df[x].values.min())
			x_max = ceil(df[x].values.max()+1)

			if(self.x_min==None):
				self.x_min=x_min
			else:
				self.x_min=min(self.x_min,x_min)

			if(self.x_max==None):
				self.x_max=x_max
			else:
				self.x_max=max(self.x_max,x_max)

			x_division_value = (self.x_max - self.x_min) // 10
			if(x_division_value==0):
				x_division_value = 1
			self.a.set(xticks=list(range(self.x_min,self.x_max,x_division_value)))

		if ('coded_'+y) in df.columns:
			
			y_ticks = df['coded_'+y].values
			self.a.set(yticks=y_ticks, yticklabels=df[y])

		else:
			y_min = floor(df[y].values.min())
			y_max = ceil(df[y].values.max()+1)

			if(self.y_min==None):
				self.y_min=y_min
			else:
				self.y_min=min(self.y_min,y_min)

			if(self.y_max==None):
				self.y_max=y_max
			else:
				self.y_max=max(self.y_max,y_max)

			y_division_value = (self.y_max - self.y_min) // 10
			if(y_division_value==0):
				y_division_value = 1
			self.a.set(yticks=list(range(self.y_min,self.y_max,y_division_value)))



		if ('coded_'+z) in df.columns:
			
			z_ticks = df['coded_'+z].values
			self.a.set(zticks=z_ticks, zticklabels=df[z])

		else:
			z_min = floor(df[z].values.min())
			z_max = ceil(df[z].values.max()+1)

			if(self.z_min==None):
				self.z_min=z_min
			else:
				self.z_min=min(self.z_min,z_min)

			if(self.z_max==None):
				self.z_max=z_max
			else:
				self.z_max=max(self.z_max,z_max)

			z_division_value = (self.z_max - self.z_min) // 10
			if(z_division_value==0):
				z_division_value = 1
			self.a.set(zticks=list(range(self.z_min,self.z_max,z_division_value)))


		scatter_proxy_shape = plt.Rectangle((0, 0), 1, 1, fc=color)
		scatter_proxy_name = label
		self.legend_proxies_shape.append(scatter_proxy_shape)
		self.legend_proxies_name.append(scatter_proxy_name)
		self.a.legend(self.legend_proxies_shape,self.legend_proxies_name)


	def add_text(self,text_content=None):

		self.a.text(0.5, 0.5, text_content, horizontalalignment='center',
			verticalalignment='center', transform=self.a.transAxes)



def main():
	test_df = pd.DataFrame({'name':['sigmod','sigmod','bbb','ccc'],'year':['2013','2014','2016','2018'],'pubcount':[5,4,3,10]})

	data_type_dict = {'name': 'str', 'venue': 'str', 'year': 'numeric', 'pubcount': 'numeric'}

	plotter = Plotter(data_convert_dict=data_type_dict,mode='3D')
	threeD_df = test_df[['name','year','pubcount']]
	plotter.plot_3D_scatter(threeD_df,x='name',y='year',z='pubcount')
	plt.show()


	plotter_2 = Plotter(data_convert_dict=data_type_dict)
	twoD_df = test_df.loc[test_df['name']!='bbb'][['name','pubcount']]
	question_df = test_df.loc[test_df['name']=='bbb']
	twoD_question_df = question_df[['name','pubcount']]
	plotter_2.plot_2D_scatter(twoD_question_df,x='name',y='pubcount',color='r',marker='P',size=100)
	plotter_2.plot_2D_scatter(twoD_df,x='name',y='pubcount')
	plt.show()


	plotter_3 = Plotter(data_convert_dict=data_type_dict,mode='2D')
	twoD_df_1 = test_df[['year','pubcount']]
	question_df_1 = test_df.loc[test_df['name']=='bbb']
	explanation_df_1 = test_df.loc[test_df['name']=='ccc']
	twoD_question_df_1 = question_df[['year','pubcount']]
	twoD_explanation_df_1 = explanation_df_1[['year','pubcount']]
	plotter_3.plot_2D_scatter(twoD_question_df_1,x='year',y='pubcount',color='r',marker='P',size=250,zorder=1)
	plotter_3.plot_2D_scatter(twoD_explanation_df_1,x='year',y='pubcount',color='g',marker='*',size=250,zorder=2)
	plotter_3.plot_2D_scatter(twoD_df_1,x='year',y='pubcount',zorder=0)
	plt.show()


if __name__=='__main__':
	main()














"""
User query template frame
"""
import tkinter as tk
from tkinter import *
from tkinter import ttk
import pandas as pd

class User_Query_Frame:

	def __init__(self,conn=None,table_dict = None,parent=None):

		self.table_dict = table_dict
		self.conn = conn
		self.parent = parent
		self.attr_list = None
		self.user_query = None

		self.query_frame=ttk.Frame(self.parent,padding=(3,3,12,12))
		self.parent.columnconfigure(0, weight=1)
		self.parent.rowconfigure(0, weight=1)
		self.query_frame.grid(column=0,row=0,sticky='nsew')
		self.query_frame.rowconfigure(0,weight=3)
		self.query_frame.rowconfigure(1,weight=1)
		self.query_frame.rowconfigure(2,weight=3)
		self.query_frame.rowconfigure(3,weight=3)
		self.query_frame.rowconfigure(4,weight=3)
		self.query_frame.rowconfigure(5,weight=3)

		self.query_frame.columnconfigure(0,weight=3)
		self.query_frame.columnconfigure(1,weight=3)
		self.query_frame.columnconfigure(2,weight=3)
		self.query_frame.columnconfigure(3,weight=3)
		self.query_frame.columnconfigure(4,weight=3)
		self.query_frame.columnconfigure(5,weight=3)

		self.select_frame = Frame(self.query_frame)
		self.select_frame.grid(row=2,column=0,columnspan=6,sticky='nsew')

		self.select_frame.columnconfigure(0,weight=4)
		self.select_frame.columnconfigure(1,weight=6)
		self.select_frame.columnconfigure(2,weight=6)
		self.select_frame.columnconfigure(3,weight=4)
		self.select_frame.columnconfigure(4,weight=1)
		self.select_frame.columnconfigure(5,weight=4)
		self.select_frame.columnconfigure(6,weight=1)
		self.select_frame.columnconfigure(7,weight=1)
		self.select_frame.columnconfigure(8,weight=4)

		self.from_where = StringVar()
		self.select_clause = StringVar()

#************************************************* Choose Table****************************************************************

		self.table_label = Label(self.query_frame,text='Choose Table:',font=('New Times Roman Bold',12))
		self.table_label.grid(row=0,column=1,sticky='nsew')
		self.table = ttk.Combobox(self.query_frame,width=15,state='readonly')
		self.table.bind('<<ComboboxSelected>>', self.table_decide_attrs)
		self.table['values'] = list(self.table_dict.keys())
		self.table.grid(row=0,column=2,sticky='ew')

#*************************************************  SELECT  ******************************************************************
		self.agg_label = Label(self.select_frame,text='Aggregate',font=('New Times Roman Bold',10))
		self.agg_label.grid(row=1,column=3,sticky='e')
		self.agg_label = Label(self.select_frame,text='Over',font=('New Times Roman Bold',10))
		self.agg_label.grid(row=1,column=5,sticky='ew')

		self.select_label = Label(self.select_frame,text="SELECT",font=('New Times Roman Bold',12))
		self.select_label.grid(row=2,column=0,sticky='ew')
		self.select_content = Label(self.select_frame, textvariable=self.select_clause,font=('New Times Roman Bold',12),wraplength=400)
		self.select_content.grid(row=2,column=1,columnspan=2,sticky='ew')
		self.agg_type = ttk.Combobox(self.select_frame,width=5,state='readonly')
		self.agg_type.grid(row=2,column=3,sticky=E)
		self.first_bracket = Label(self.select_frame,text='(')
		self.first_bracket.grid(row=2,column=4,sticky=E)
		self.agg_attr = ttk.Combobox(self.select_frame,width=8,state='readonly')
		self.agg_attr.bind('<<ComboboxSelected>>', self.update_query_frame)
		self.agg_attr.grid(row=2,column=5,sticky='ew')
		self.second_bracket = Label(self.select_frame,text=')')
		self.second_bracket.grid(row=2,column=6,sticky=W)
		self.as_label = Label(self.select_frame,text=' AS ',font=('New Times Roman Bold',12))
		self.as_label.grid(row=2,column=7,sticky='we')
		self.alias = Entry(self.select_frame,bg="white",width=2)
		self.alias.grid(row=2,column=8,sticky='we')

#*************************************************   FROM   ********************************************************************     
		self.from_table = Label(self.query_frame,text="FROM",font=('New Times Roman Bold',12))
		self.from_table.grid(row=3,column=0,sticky='ew')
		self.from_table = Label(self.query_frame,textvariable=self.from_where,font=('New Times Roman Bold',12))
		self.from_table.grid(row=3,column=1,sticky='ew')

#*****************************************************GROUP BY ******************************************************************		
		self.group_by_label = Label(self.query_frame,text="GROUP BY",font=('New Times Roman Bold',12))
		self.group_by_label.grid(column=0,row=4)

		self.group1 = ttk.Combobox(self.query_frame,width=13,state='readonly')
		self.group1.grid(column=1,row=4)
		self.group1.bind('<<ComboboxSelected>>', self.update_query_frame)

		self.group2 = ttk.Combobox(self.query_frame,width=13,state='readonly')
		self.group2.grid(column=2,row=4)
		self.group2.bind('<<ComboboxSelected>>', self.update_query_frame)

		self.group3 = ttk.Combobox(self.query_frame,width=13,state='readonly')
		self.group3.grid(column=3,row=4)
		self.group3.bind('<<ComboboxSelected>>', self.update_query_frame)

		self.group4 = ttk.Combobox(self.query_frame,width=13,state='readonly')
		self.group4.grid(column=4,row=4)
		self.group4.bind('<<ComboboxSelected>>', self.update_query_frame)

		self.group5 = ttk.Combobox(self.query_frame,width=13,state='readonly')
		self.group5.grid(column=5,row=4)
		self.group5.bind('<<ComboboxSelected>>', self.update_query_frame)

#******************************************************Functions*****************************************************************

	def table_decide_attrs(self,event):
		self.attr_list= self.table_dict[self.table.get()]
		self.from_where.set(self.table.get())
		self.shown_attrs = self.attr_list
		self.shown_attrs.insert(0,' ')
		self.shown_attrs.insert(1,'*')
		self.shown_attrs.sort()
		self.group1.set(' ')
		self.group2.set(' ')
		self.group3.set(' ')
		self.group4.set(' ')
		self.group5.set(' ')
		self.agg_type.set(' ')
		self.agg_attr.set(' ')
		self.group1['values'] = self.shown_attrs
		self.group2['values'] = self.shown_attrs
		self.group3['values'] = self.shown_attrs
		self.group4['values'] = self.shown_attrs
		self.group5['values'] = self.shown_attrs
		self.agg_attr['values'] = self.shown_attrs

		table_name = self.table.get()
		q = "select distinct substring( agg from '(sum|max|avg|min|count)') as agg from dev." +str(table_name)+'_global'
		agg_df = pd.read_sql(q,self.conn)
		agg_list = agg_df['agg'].values.tolist()
		self.agg_type['values'] = agg_list

	def update_query_frame(self,event):
		g1 = self.group1.get()
		g2 = self.group2.get()
		g3 = self.group3.get()
		g4 = self.group4.get()
		g5 = self.group5.get()
		attr = self.agg_attr.get()
		self.chosen_attrs = [g1,g2,g3,g4,g5,attr]
		self.shown_attrs = list(set(self.attr_list) - set(list(filter(None, self.chosen_attrs))))
		self.selected_attrs = list(filter(lambda x :' ' not in x,self.chosen_attrs ))
		if(self.shown_attrs == []):
			self.shown_attrs = [' ']
		if(self.shown_attrs[0]!=' '):
			self.shown_attrs.insert(0,' ')
		self.shown_attrs.sort()
		self.group1['values'] = self.shown_attrs
		self.group2['values'] = self.shown_attrs
		self.group3['values'] = self.shown_attrs
		self.group4['values'] = self.shown_attrs
		self.group5['values'] = self.shown_attrs
		self.agg_attr['values'] = self.shown_attrs
		if attr in self.selected_attrs:
			self.selected_attrs.remove(attr)
		self.select_clause.set(','.join(self.selected_attrs))

	def get_query(self):
		query_group_str = self.select_clause.get()
		query_group_list = query_group_str.split(',')
		query_group_list.sort()
		sorted_query_group = ','.join(query_group_list)
		alias_name = self.alias.get().lower()
		table_name = self.from_where.get()
		agg_function=self.agg_type.get()+'('+self.agg_attr.get()+')'

		user_agg =self.agg_type.get()+'_'+self.agg_attr.get()

		query = "SELECT " +self.select_clause.get() + ","+agg_function+ " as " + alias_name+\
		"\nFROM " + table_name+\
		"\nGROUP BY " + query_group_str+";"

		return query,sorted_query_group,agg_function,user_agg,alias_name,table_name



if __name__ == '__main__':
	root = Tk()
	uqf = User_Query_Frame(parent=root)
	root.title('CAPE')
	root.mainloop()



		







		




from itertools import combinations,permutations

def add_rollup(dic,group,prefix,division,d_index,redundant):
    for k in range(d_index,prefix,-1):
        if division and division>=k:
                division=None
        if not division:
            for i in range(k):
                key=tuple(group[:i])
                if key not in dic:
                    dic[key]=set()
                v=group[i:k]
                if v not in dic[key]:
                    dic[key].add(v)
                else:
                    if key not in redundant:
                        redundant[key]=set()
                    redundant[key].add(v)
        else:
            key=tuple(group[:division])
            if key not in dic:
                dic[key]=set()
            v=group[division:k]
            if v not in dic[key]:
                dic[key].add(v)
            else:
                if key not in redundant:
                    redundant[key]=set()
                redundant[key].add(v)

#when there are n numbers and group, return the next available for group[-1]
def nextnum(group,n):
    res=(group[-1]+1)%n
    if len(group)>1:
        while res in group[:-1]:
            res=(res+1)%n
    return res

#find the prefix that we want to do regression for this group. If a prefix
#will be taken care of by a different grouping, then not do it.            
def findpre(group,d_index,n):
    for k in range(d_index,0,-1):
        if group[k]!=nextnum(group[:k],n):
            return k
    return 0

def main():
    n=8
    l=[i for i in range(n)]
    
    all_group={}
    for i in range(4,0,-1):
        groups=combinations(l,i)
        for group in groups:
            for j in range(len(group),0,-1):
                vs=combinations(group,j)
                for v in vs:
                    f=tuple(sorted([f for f in group if f not in v]))
                    if f not in all_group:
                        all_group[f]=set()
                    all_group[f].add(v)
    
    perm_group={}
    redundant={}
    combs=combinations(l,min(4,len(l)))
    for comb in combs:
        perms=permutations(comb,len(comb))
        for perm in perms:
            decrease=0
            d_index=None
            division=None
            for i in range(1,len(perm)):
                if perm[i-1]>perm[i]:
                    decrease+=1
                    if decrease==1:
                        division=i
                if decrease==2:
                    d_index=i #perm[:d_index] will decrease at most once
                    
            if not d_index:
                d_index=len(perm)
                pre=findpre(perm,d_index-1,n)
            else:
                pre=findpre(perm,d_index,n)
            
            if pre==d_index:
                continue
            else:
                add_rollup(perm_group,perm,pre,division,d_index,redundant)
            
    dif={}
    for key in all_group:
        if key not in perm_group:
            dif[key]=all_group[key]
        else:
            for v in all_group[key]:
                if v not in perm_group[key]:
                    if key not in dif:
                        dif[key]=set()
                    dif[key].add(v)
    
    print(dif)
    print(redundant)

if __name__=="__main__":
    main()
